# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

__all__ = ['expr', 'RqlQuery', 'ReQLEncoder', 'ReQLDecoder', 'Repl']


import base64
import binascii
import collections
import datetime
import json
import threading

from rethinkdb import ql2_pb2
from rethinkdb.errors import QueryPrinter, ReqlDriverCompileError, ReqlDriverError, T

P_TERM = ql2_pb2.Term.TermType

try:
    unicode
except NameError:
    unicode = str

try:
    xrange
except NameError:
    xrange = range


def dict_items(dictionary):
    return list(dictionary.items())


class Repl(object):
    thread_data = threading.local()
    repl_active = False

    @classmethod
    def get(cls):
        if 'repl' in cls.thread_data.__dict__:
            return cls.thread_data.repl
        else:
            return None

    @classmethod
    def set(cls, conn):
        cls.thread_data.repl = conn
        cls.repl_active = True

    @classmethod
    def clear(cls):
        if 'repl' in cls.thread_data.__dict__:
            del cls.thread_data.repl
        cls.repl_active = False

# This is both an external function and one used extensively
# internally to convert coerce python values to RQL types


def expr(val, nesting_depth=20):
    '''
        Convert a Python primitive into a RQL primitive value
    '''
    if not isinstance(nesting_depth, int):
        raise ReqlDriverCompileError("Second argument to `r.expr` must be a number.")

    if nesting_depth <= 0:
        raise ReqlDriverCompileError("Nesting depth limit exceeded.")

    if isinstance(val, RqlQuery):
        return val
    elif isinstance(val, collections.Callable):
        return Func(val)
    elif isinstance(val, (datetime.datetime, datetime.date)):
        if not hasattr(val, 'tzinfo') or not val.tzinfo:
            raise ReqlDriverCompileError("""Cannot convert %s to ReQL time object
            without timezone information. You can add timezone information with
            the third party module \"pytz\" or by constructing ReQL compatible
            timezone values with r.make_timezone(\"[+-]HH:MM\"). Alternatively,
            use one of ReQL's bultin time constructors, r.now, r.time,
            or r.iso8601.
            """ % (type(val).__name__))
        return ISO8601(val.isoformat())
    elif isinstance(val, RqlBinary):
        return Binary(val)
    elif isinstance(val, (str, unicode)):
        return Datum(val)
    elif isinstance(val, bytes):
        return Binary(val)
    elif isinstance(val, collections.Mapping):
        # MakeObj doesn't take the dict as a keyword args to avoid
        # conflicting with the `self` parameter.
        obj = {}
        for k, v in dict_items(val):
            obj[k] = expr(v, nesting_depth - 1)
        return MakeObj(obj)
    elif isinstance(val, collections.Iterable):
        val = [expr(v, nesting_depth - 1) for v in val]
        return MakeArray(*val)
    else:
        return Datum(val)


class RqlQuery(object):
    # Instantiate this AST node with the given pos and opt args
    def __init__(self, *args, **optargs):
        self._args = [expr(e) for e in args]

        self.optargs = {}
        for key, value in dict_items(optargs):
            self.optargs[key] = expr(value)

    # Send this query to the server to be executed
    def run(self, c=None, **global_optargs):
        if c is None:
            c = Repl.get()
            if c is None:
                if Repl.repl_active:
                    raise ReqlDriverError(
                        "RqlQuery.run must be given a connection to run on. A default connection has been set with "
                        "`repl()` on another thread, but not this one."
                    )
                else:
                    raise ReqlDriverError("RqlQuery.run must be given a connection to run on.")

        return c._start(self, **global_optargs)

    def __str__(self):
        printer = QueryPrinter(self)
        return printer.print_query()

    def __repr__(self):
        return "<RqlQuery instance: %s >" % str(self)

    # Compile this query to a json-serializable object
    def build(self):
        res = [self.term_type, self._args]
        if len(self.optargs) > 0:
            res.append(self.optargs)
        return res

    # The following are all operators and methods that operate on
    # Rql queries to build up more complex operations

    # Comparison operators
    def __eq__(self, other):
        return Eq(self, other)

    def __ne__(self, other):
        return Ne(self, other)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    # Numeric operators
    def __invert__(self):
        return Not(self)

    def __add__(self, other):
        return Add(self, other)

    def __radd__(self, other):
        return Add(other, self)

    def __sub__(self, other):
        return Sub(self, other)

    def __rsub__(self, other):
        return Sub(other, self)

    def __mul__(self, other):
        return Mul(self, other)

    def __rmul__(self, other):
        return Mul(other, self)

    def __div__(self, other):
        return Div(self, other)

    def __rdiv__(self, other):
        return Div(other, self)

    def __truediv__(self, other):
        return Div(self, other)

    def __rtruediv__(self, other):
        return Div(other, self)

    def __mod__(self, other):
        return Mod(self, other)

    def __rmod__(self, other):
        return Mod(other, self)

    def __and__(self, other):
        query = And(self, other)
        query.set_infix()
        return query

    def __rand__(self, other):
        query = And(other, self)
        query.set_infix()
        return query

    def __or__(self, other):
        query = Or(self, other)
        query.set_infix()
        return query

    def __ror__(self, other):
        query = Or(other, self)
        query.set_infix()
        return query

    # Non-operator versions of the above

    def eq(self, *args):
        return Eq(self, *args)

    def ne(self, *args):
        return Ne(self, *args)

    def lt(self, *args):
        return Lt(self, *args)

    def le(self, *args):
        return Le(self, *args)

    def gt(self, *args):
        return Gt(self, *args)

    def ge(self, *args):
        return Ge(self, *args)

    def add(self, *args):
        return Add(self, *args)

    def sub(self, *args):
        return Sub(self, *args)

    def mul(self, *args):
        return Mul(self, *args)

    def div(self, *args):
        return Div(self, *args)

    def mod(self, *args):
        return Mod(self, *args)

    def bit_and(self, *args):
        return BitAnd(self, *args)

    def bit_or(self, *args):
        return BitOr(self, *args)

    def bit_xor(self, *args):
        return BitXor(self, *args)

    def bit_not(self, *args):
        return BitNot(self, *args)

    def bit_sal(self, *args):
        return BitSal(self, *args)

    def bit_sar(self, *args):
        return BitSar(self, *args)

    def floor(self, *args):
        return Floor(self, *args)

    def ceil(self, *args):
        return Ceil(self, *args)

    def round(self, *args):
        return Round(self, *args)

    def and_(self, *args):
        return And(self, *args)

    def or_(self, *args):
        return Or(self, *args)

    def not_(self, *args):
        return Not(self, *args)

    # N.B. Cannot use 'in' operator because it must return a boolean
    def contains(self, *args):
        return Contains(self, *[func_wrap(arg) for arg in args])

    def has_fields(self, *args):
        return HasFields(self, *args)

    def with_fields(self, *args):
        return WithFields(self, *args)

    def keys(self, *args):
        return Keys(self, *args)

    def values(self, *args):
        return Values(self, *args)

    def changes(self, *args, **kwargs):
        return Changes(self, *args, **kwargs)

    # Polymorphic object/sequence operations
    def pluck(self, *args):
        return Pluck(self, *args)

    def without(self, *args):
        return Without(self, *args)

    def do(self, *args):
        return FunCall(self, *args)

    def default(self, *args):
        return Default(self, *args)

    def update(self, *args, **kwargs):
        return Update(self, *[func_wrap(arg) for arg in args], **kwargs)

    def replace(self, *args, **kwargs):
        return Replace(self, *[func_wrap(arg) for arg in args], **kwargs)

    def delete(self, *args, **kwargs):
        return Delete(self, *args, **kwargs)

    # Rql type inspection
    def coerce_to(self, *args):
        return CoerceTo(self, *args)

    def ungroup(self, *args):
        return Ungroup(self, *args)

    def type_of(self, *args):
        return TypeOf(self, *args)

    def merge(self, *args):
        return Merge(self, *[func_wrap(arg) for arg in args])

    def append(self, *args):
        return Append(self, *args)

    def prepend(self, *args):
        return Prepend(self, *args)

    def difference(self, *args):
        return Difference(self, *args)

    def set_insert(self, *args):
        return SetInsert(self, *args)

    def set_union(self, *args):
        return SetUnion(self, *args)

    def set_intersection(self, *args):
        return SetIntersection(self, *args)

    def set_difference(self, *args):
        return SetDifference(self, *args)

    # Operator used for get attr / nth / slice. Non-operator versions below
    # in cases of ambiguity
    def __getitem__(self, index):
        if isinstance(index, slice):
            if index.stop:
                return Slice(self, index.start or 0, index.stop,
                             bracket_operator=True)
            else:
                return Slice(self, index.start or 0, -1,
                             right_bound='closed', bracket_operator=True)
        else:
            return Bracket(self, index, bracket_operator=True)

    def __iter__(*args, **kwargs):
        raise ReqlDriverError(
            "__iter__ called on an RqlQuery object.\n"
            "To iterate over the results of a query, call run first.\n"
            "To iterate inside a query, use map or for_each.")

    def get_field(self, *args):
        return GetField(self, *args)

    def nth(self, *args):
        return Nth(self, *args)

    def to_json(self, *args):
        return ToJsonString(self, *args)

    def to_json_string(self, *args):
        return ToJsonString(self, *args)

    def match(self, *args):
        return Match(self, *args)

    def split(self, *args):
        return Split(self, *args)

    def upcase(self, *args):
        return Upcase(self, *args)

    def downcase(self, *args):
        return Downcase(self, *args)

    def is_empty(self, *args):
        return IsEmpty(self, *args)

    def offsets_of(self, *args):
        return OffsetsOf(self, *[func_wrap(arg) for arg in args])

    def slice(self, *args, **kwargs):
        return Slice(self, *args, **kwargs)

    def skip(self, *args):
        return Skip(self, *args)

    def limit(self, *args):
        return Limit(self, *args)

    def reduce(self, *args):
        return Reduce(self, *[func_wrap(arg) for arg in args])

    def sum(self, *args):
        return Sum(self, *[func_wrap(arg) for arg in args])

    def avg(self, *args):
        return Avg(self, *[func_wrap(arg) for arg in args])

    def min(self, *args, **kwargs):
        return Min(self, *[func_wrap(arg) for arg in args], **kwargs)

    def max(self, *args, **kwargs):
        return Max(self, *[func_wrap(arg) for arg in args], **kwargs)

    def map(self, *args):
        if len(args) > 0:
            # `func_wrap` only the last argument
            return Map(self, *(args[:-1] + (func_wrap(args[-1]), )))
        else:
            return Map(self)

    def fold(self, *args, **kwargs):
        if len(args) > 0:
            # `func_wrap` only the last argument before optional arguments
            # Also `func_wrap` keyword arguments

            # Nice syntax not supported by python2.6
            kwfuncargs = {}
            for arg_name in kwargs:
                kwfuncargs[arg_name] = func_wrap(kwargs[arg_name])
            return Fold(self, *(args[:-1] + (func_wrap(args[-1]), )), **kwfuncargs)
        else:
            return Fold(self)

    def filter(self, *args, **kwargs):
        return Filter(self, *[func_wrap(arg) for arg in args], **kwargs)

    def concat_map(self, *args):
        return ConcatMap(self, *[func_wrap(arg) for arg in args])

    def order_by(self, *args, **kwargs):
        args = [arg if isinstance(arg, (Asc, Desc)) else func_wrap(arg)
                for arg in args]
        return OrderBy(self, *args, **kwargs)

    def between(self, *args, **kwargs):
        return Between(self, *args, **kwargs)

    def distinct(self, *args, **kwargs):
        return Distinct(self, *args, **kwargs)

    # NB: Can't overload __len__ because Python doesn't
    #     allow us to return a non-integer
    def count(self, *args):
        return Count(self, *[func_wrap(arg) for arg in args])

    def union(self, *args, **kwargs):
        func_kwargs = {}
        for key in kwargs:
            func_kwargs[key] = func_wrap(kwargs[key])
        return Union(self, *args, **func_kwargs)

    def inner_join(self, *args):
        return InnerJoin(self, *args)

    def outer_join(self, *args):
        return OuterJoin(self, *args)

    def eq_join(self, *args, **kwargs):
        return EqJoin(self, *[func_wrap(arg) for arg in args], **kwargs)

    def zip(self, *args):
        return Zip(self, *args)

    def group(self, *args, **kwargs):
        return Group(self, *[func_wrap(arg) for arg in args], **kwargs)

    def branch(self, *args):
        return Branch(self, *args)

    def for_each(self, *args):
        return ForEach(self, *[func_wrap(arg) for arg in args])

    def info(self, *args):
        return Info(self, *args)

    # Array only operations
    def insert_at(self, *args):
        return InsertAt(self, *args)

    def splice_at(self, *args):
        return SpliceAt(self, *args)

    def delete_at(self, *args):
        return DeleteAt(self, *args)

    def change_at(self, *args):
        return ChangeAt(self, *args)

    def sample(self, *args):
        return Sample(self, *args)

    # Time support

    def to_iso8601(self, *args):
        return ToISO8601(self, *args)

    def to_epoch_time(self, *args):
        return ToEpochTime(self, *args)

    def during(self, *args, **kwargs):
        return During(self, *args, **kwargs)

    def date(self, *args):
        return Date(self, *args)

    def time_of_day(self, *args):
        return TimeOfDay(self, *args)

    def timezone(self, *args):
        return Timezone(self, *args)

    def year(self, *args):
        return Year(self, *args)

    def month(self, *args):
        return Month(self, *args)

    def day(self, *args):
        return Day(self, *args)

    def day_of_week(self, *args):
        return DayOfWeek(self, *args)

    def day_of_year(self, *args):
        return DayOfYear(self, *args)

    def hours(self, *args):
        return Hours(self, *args)

    def minutes(self, *args):
        return Minutes(self, *args)

    def seconds(self, *args):
        return Seconds(self, *args)

    def in_timezone(self, *args):
        return InTimezone(self, *args)

    # Geospatial support
    def to_geojson(self, *args):
        return ToGeoJson(self, *args)

    def distance(self, *args, **kwargs):
        return Distance(self, *args, **kwargs)

    def intersects(self, *args):
        return Intersects(self, *args)

    def includes(self, *args):
        return Includes(self, *args)

    def fill(self, *args):
        return Fill(self, *args)

    def polygon_sub(self, *args):
        return PolygonSub(self, *args)


# These classes define how nodes are printed by overloading `compose`
def needs_wrap(arg):
    return isinstance(arg, (Datum, MakeArray, MakeObj))


class RqlBoolOperQuery(RqlQuery):
    def __init__(self, *args, **optargs):
        self.infix = False
        RqlQuery.__init__(self, *args, **optargs)

    def set_infix(self):
        self.infix = True

    def compose(self, args, optargs):
        t_args = [T('r.expr(', args[i], ')')
                  if needs_wrap(self._args[i]) else args[i]
                  for i in xrange(len(args))]

        if self.infix:
            return T('(', T(*t_args, intsp=[' ', self.statement_infix, ' ']), ')')
        else:
            return T('r.', self.statement, '(', T(*t_args, intsp=', '), ')')


class RqlBiOperQuery(RqlQuery):
    def compose(self, args, optargs):
        t_args = [T('r.expr(', args[i], ')')
                  if needs_wrap(self._args[i]) else args[i]
                  for i in xrange(len(args))]
        return T('(', T(*t_args, intsp=[' ', self.statement, ' ']), ')')


class RqlBiCompareOperQuery(RqlBiOperQuery):
    def __init__(self, *args, **optargs):
        RqlBiOperQuery.__init__(self, *args, **optargs)

        for arg in args:
            try:
                if arg.infix:
                    err = (
                        "Calling '%s' on result of infix bitwise operator:\n"
                        "%s.\n"
                        "This is almost always a precedence error.\n"
                        "Note that `a < b | b < c` <==> `a < (b | b) < c`.\n"
                        "If you really want this behavior, use `.or_` or "
                        "`.and_` instead.")
                    raise ReqlDriverCompileError(err %
                                                 (self.statement,
                                                  QueryPrinter(self).print_query()))
            except AttributeError:
                pass  # No infix attribute, so not possible to be an infix bool operator


class RqlTopLevelQuery(RqlQuery):
    def compose(self, args, optargs):
        args.extend([T(key, '=', value) for key, value in dict_items(optargs)])
        return T('r.', self.statement, '(', T(*(args), intsp=', '), ')')


class RqlMethodQuery(RqlQuery):
    def compose(self, args, optargs):
        if len(args) == 0:
            return T('r.', self.statement, '()')

        if needs_wrap(self._args[0]):
            args[0] = T('r.expr(', args[0], ')')

        restargs = args[1:]
        restargs.extend([T(k, '=', v) for k, v in dict_items(optargs)])
        restargs = T(*restargs, intsp=', ')

        return T(args[0], '.', self.statement, '(', restargs, ')')


class RqlBracketQuery(RqlMethodQuery):
    def __init__(self, *args, **optargs):
        if 'bracket_operator' in optargs:
            self.bracket_operator = optargs['bracket_operator']
            del optargs['bracket_operator']
        else:
            self.bracket_operator = False

        RqlMethodQuery.__init__(self, *args, **optargs)

    def compose(self, args, optargs):
        if self.bracket_operator:
            if needs_wrap(self._args[0]):
                args[0] = T('r.expr(', args[0], ')')
            return T(args[0], '[', T(*args[1:], intsp=[',']), ']')
        else:
            return RqlMethodQuery.compose(self, args, optargs)


class RqlTzinfo(datetime.tzinfo):
    def __init__(self, offsetstr):
        hours, minutes = map(int, offsetstr.split(':'))

        self.offsetstr = offsetstr
        self.delta = datetime.timedelta(hours=hours, minutes=minutes)

    def __getinitargs__(self):
        return (self.offsetstr,)

    def __copy__(self):
        return RqlTzinfo(self.offsetstr)

    def __deepcopy__(self, memo):
        return RqlTzinfo(self.offsetstr)

    def utcoffset(self, dt):
        return self.delta

    def tzname(self, dt):
        return self.offsetstr

    def dst(self, dt):
        return datetime.timedelta(0)


# Python only allows immutable built-in types to be hashed, such as
# for keys in a dict This means we can't use lists or dicts as keys in
# grouped data objects, so we convert them to tuples and frozensets,
# respectively.  This may make it a little harder for users to work
# with converted grouped data, unless they do a simple iteration over
# the result
def recursively_make_hashable(obj):
    if isinstance(obj, list):
        return tuple([recursively_make_hashable(i) for i in obj])
    elif isinstance(obj, dict):
        return frozenset([(k, recursively_make_hashable(v))
                          for k, v in dict_items(obj)])
    return obj


class ReQLEncoder(json.JSONEncoder):
    '''
        Default JSONEncoder subclass to handle query conversion.
    '''

    def __init__(self):
        json.JSONEncoder.__init__(
            self, ensure_ascii=False, allow_nan=False, check_circular=False, separators=(',', ':')
        )

    def default(self, obj):
        if isinstance(obj, RqlQuery):
            return obj.build()
        return json.JSONEncoder.default(self, obj)


class ReQLDecoder(json.JSONDecoder):
    '''
        Default JSONDecoder subclass to handle pseudo-type conversion.
    '''

    def __init__(self, reql_format_opts=None):
        json.JSONDecoder.__init__(self, object_hook=self.convert_pseudotype)
        self.reql_format_opts = reql_format_opts or {}

    def convert_time(self, obj):
        if 'epoch_time' not in obj:
            raise ReqlDriverError(('pseudo-type TIME object %s does not ' +
                                   'have expected field "epoch_time".')
                                  % json.dumps(obj))

        if 'timezone' in obj:
            return datetime.datetime.fromtimestamp(obj['epoch_time'],
                                                   RqlTzinfo(obj['timezone']))
        else:
            return datetime.datetime.utcfromtimestamp(obj['epoch_time'])

    @staticmethod
    def convert_grouped_data(obj):
        if 'data' not in obj:
            raise ReqlDriverError(('pseudo-type GROUPED_DATA object' +
                                   ' %s does not have the expected field "data".')
                                  % json.dumps(obj))
        return dict([(recursively_make_hashable(k), v) for k, v in obj['data']])

    @staticmethod
    def convert_binary(obj):
        if 'data' not in obj:
            raise ReqlDriverError(('pseudo-type BINARY object %s does not have ' +
                                   'the expected field "data".')
                                  % json.dumps(obj))
        return RqlBinary(base64.b64decode(obj['data'].encode('utf-8')))

    def convert_pseudotype(self, obj):
        reql_type = obj.get('$reql_type$')
        if reql_type is not None:
            if reql_type == 'TIME':
                time_format = self.reql_format_opts.get('time_format')
                if time_format is None or time_format == 'native':
                    # Convert to native python datetime object
                    return self.convert_time(obj)
                elif time_format != 'raw':
                    raise ReqlDriverError("Unknown time_format run option \"%s\"."
                                          % time_format)
            elif reql_type == 'GROUPED_DATA':
                group_format = self.reql_format_opts.get('group_format')
                if group_format is None or group_format == 'native':
                    return self.convert_grouped_data(obj)
                elif group_format != 'raw':
                    raise ReqlDriverError("Unknown group_format run option \"%s\"."
                                          % group_format)
            elif reql_type == 'GEOMETRY':
                # No special support for this. Just return the raw object
                return obj
            elif reql_type == 'BINARY':
                binary_format = self.reql_format_opts.get('binary_format')
                if binary_format is None or binary_format == 'native':
                    return self.convert_binary(obj)
                elif binary_format != 'raw':
                    raise ReqlDriverError("Unknown binary_format run option \"%s\"."
                                          % binary_format)
            else:
                raise ReqlDriverError("Unknown pseudo-type %s" % reql_type)
        # If there was no pseudotype, or the relevant format is raw, return
        # the original object
        return obj


# This class handles the conversion of RQL terminal types in both directions
# Going to the server though it does not support R_ARRAY or R_OBJECT as those
# are alternately handled by the MakeArray and MakeObject nodes. Why do this?
# MakeArray and MakeObject are more flexible, allowing us to construct array
# and object expressions from nested RQL expressions. Constructing pure
# R_ARRAYs and R_OBJECTs would require verifying that at all nested levels
# our arrays and objects are composed only of basic types.
class Datum(RqlQuery):
    def __init__(self, val):
        super(Datum, self).__init__()
        self.data = val

    def build(self):
        return self.data

    def compose(self, args, optargs):
        return repr(self.data)


class MakeArray(RqlQuery):
    term_type = P_TERM.MAKE_ARRAY

    def compose(self, args, optargs):
        return T('[', T(*args, intsp=', '), ']')


class MakeObj(RqlQuery):
    term_type = P_TERM.MAKE_OBJ

    # We cannot inherit from RqlQuery because of potential conflicts with
    # the `self` parameter. This is not a problem for other RqlQuery sub-
    # classes unless we add a 'self' optional argument to one of them.
    # TODO: @gabor-boros Figure out is the above still an issue or not.
    def __init__(self, obj_dict):
        super(MakeObj, self).__init__()
        for key, value in dict_items(obj_dict):
            if not isinstance(key, (str, unicode)):
                raise ReqlDriverCompileError("Object keys must be strings.")
            self.optargs[key] = expr(value)

    def build(self):
        return self.optargs

    def compose(self, args, optargs):
        return T('r.expr({', T(*[T(repr(key), ': ', value)
                                 for key, value in dict_items(optargs)],
                               intsp=', '), '})')


class Var(RqlQuery):
    term_type = P_TERM.VAR

    def compose(self, args, optargs):
        return 'var_' + args[0]


class JavaScript(RqlTopLevelQuery):
    term_type = P_TERM.JAVASCRIPT
    statement = "js"


class Http(RqlTopLevelQuery):
    term_type = P_TERM.HTTP
    statement = "http"


class UserError(RqlTopLevelQuery):
    term_type = P_TERM.ERROR
    statement = "error"


class Random(RqlTopLevelQuery):
    term_type = P_TERM.RANDOM
    statement = "random"


class Changes(RqlMethodQuery):
    term_type = P_TERM.CHANGES
    statement = "changes"


class Default(RqlMethodQuery):
    term_type = P_TERM.DEFAULT
    statement = "default"


class ImplicitVar(RqlQuery):
    term_type = P_TERM.IMPLICIT_VAR

    def __call__(self, *args, **kwargs):
        raise TypeError("'r.row' is not callable, use 'r.row[...]' instead")

    def compose(self, args, optargs):
        return 'r.row'


class Eq(RqlBiCompareOperQuery):
    term_type = P_TERM.EQ
    statement = "=="


class Ne(RqlBiCompareOperQuery):
    term_type = P_TERM.NE
    statement = "!="


class Lt(RqlBiCompareOperQuery):
    term_type = P_TERM.LT
    statement = "<"


class Le(RqlBiCompareOperQuery):
    term_type = P_TERM.LE
    statement = "<="


class Gt(RqlBiCompareOperQuery):
    term_type = P_TERM.GT
    statement = ">"


class Ge(RqlBiCompareOperQuery):
    term_type = P_TERM.GE
    statement = ">="


class Not(RqlQuery):
    term_type = P_TERM.NOT

    def compose(self, args, optargs):
        if isinstance(self._args[0], Datum):
            args[0] = T('r.expr(', args[0], ')')
        return T('(~', args[0], ')')


class Add(RqlBiOperQuery):
    term_type = P_TERM.ADD
    statement = "+"


class Sub(RqlBiOperQuery):
    term_type = P_TERM.SUB
    statement = "-"


class Mul(RqlBiOperQuery):
    term_type = P_TERM.MUL
    statement = "*"


class Div(RqlBiOperQuery):
    term_type = P_TERM.DIV
    statement = "/"


class Mod(RqlBiOperQuery):
    term_type = P_TERM.MOD
    statement = "%"


class BitAnd(RqlBoolOperQuery):
    term_type = P_TERM.BIT_AND
    statement = "bit_and"


class BitOr(RqlBoolOperQuery):
    term_type = P_TERM.BIT_OR
    statement = "bit_or"


class BitXor(RqlBoolOperQuery):
    term_type = P_TERM.BIT_XOR
    statement = "bit_xor"


class BitNot(RqlMethodQuery):
    term_type = P_TERM.BIT_NOT
    statement = "bit_not"


class BitSal(RqlBoolOperQuery):
    term_type = P_TERM.BIT_SAL
    statement = "bit_sal"


class BitSar(RqlBoolOperQuery):
    term_type = P_TERM.BIT_SAR
    statement = "bit_sar"


class Floor(RqlMethodQuery):
    term_type = P_TERM.FLOOR
    statement = 'floor'


class Ceil(RqlMethodQuery):
    term_type = P_TERM.CEIL
    statement = 'ceil'


class Round(RqlMethodQuery):
    term_type = P_TERM.ROUND
    statement = 'round'


class Append(RqlMethodQuery):
    term_type = P_TERM.APPEND
    statement = "append"


class Prepend(RqlMethodQuery):
    term_type = P_TERM.PREPEND
    statement = "prepend"


class Difference(RqlMethodQuery):
    term_type = P_TERM.DIFFERENCE
    statement = "difference"


class SetInsert(RqlMethodQuery):
    term_type = P_TERM.SET_INSERT
    statement = "set_insert"


class SetUnion(RqlMethodQuery):
    term_type = P_TERM.SET_UNION
    statement = "set_union"


class SetIntersection(RqlMethodQuery):
    term_type = P_TERM.SET_INTERSECTION
    statement = "set_intersection"


class SetDifference(RqlMethodQuery):
    term_type = P_TERM.SET_DIFFERENCE
    statement = "set_difference"


class Slice(RqlBracketQuery):
    term_type = P_TERM.SLICE
    statement = 'slice'

    # Slice has a special bracket syntax, implemented here
    def compose(self, args, optargs):
        if self.bracket_operator:
            if needs_wrap(self._args[0]):
                args[0] = T('r.expr(', args[0], ')')
            return T(args[0], '[', args[1], ':', args[2], ']')
        else:
            return RqlBracketQuery.compose(self, args, optargs)


class Skip(RqlMethodQuery):
    term_type = P_TERM.SKIP
    statement = 'skip'


class Limit(RqlMethodQuery):
    term_type = P_TERM.LIMIT
    statement = 'limit'


class GetField(RqlBracketQuery):
    term_type = P_TERM.GET_FIELD
    statement = 'get_field'


class Bracket(RqlBracketQuery):
    term_type = P_TERM.BRACKET
    statement = 'bracket'


class Contains(RqlMethodQuery):
    term_type = P_TERM.CONTAINS
    statement = 'contains'


class HasFields(RqlMethodQuery):
    term_type = P_TERM.HAS_FIELDS
    statement = 'has_fields'


class WithFields(RqlMethodQuery):
    term_type = P_TERM.WITH_FIELDS
    statement = 'with_fields'


class Keys(RqlMethodQuery):
    term_type = P_TERM.KEYS
    statement = 'keys'


class Values(RqlMethodQuery):
    term_type = P_TERM.VALUES
    statement = 'values'


class Object(RqlMethodQuery):
    term_type = P_TERM.OBJECT
    statement = 'object'


class Pluck(RqlMethodQuery):
    term_type = P_TERM.PLUCK
    statement = 'pluck'


class Without(RqlMethodQuery):
    term_type = P_TERM.WITHOUT
    statement = 'without'


class Merge(RqlMethodQuery):
    term_type = P_TERM.MERGE
    statement = 'merge'


class Between(RqlMethodQuery):
    term_type = P_TERM.BETWEEN
    statement = 'between'


class DB(RqlTopLevelQuery):
    term_type = P_TERM.DB
    statement = 'db'

    def table_list(self, *args):
        return TableList(self, *args)

    def config(self, *args):
        return Config(self, *args)

    def wait(self, *args, **kwargs):
        return Wait(self, *args, **kwargs)

    def reconfigure(self, *args, **kwargs):
        return Reconfigure(self, *args, **kwargs)

    def rebalance(self, *args, **kwargs):
        return Rebalance(self, *args, **kwargs)

    def grant(self, *args, **kwargs):
        return Grant(self, *args, **kwargs)

    def table_create(self, *args, **kwargs):
        return TableCreate(self, *args, **kwargs)

    def table_drop(self, *args):
        return TableDrop(self, *args)

    def table(self, *args, **kwargs):
        return Table(self, *args, **kwargs)


class FunCall(RqlQuery):
    term_type = P_TERM.FUNCALL

    # This object should be constructed with arguments first, and the
    # function itself as the last parameter.  This makes it easier for
    # the places where this object is constructed.  The actual wire
    # format is function first, arguments last, so we flip them around
    # before passing it down to the base class constructor.
    def __init__(self, *args):
        if len(args) == 0:
            raise ReqlDriverCompileError("Expected 1 or more arguments but found 0.")
        args = [func_wrap(args[-1])] + list(args[:-1])
        RqlQuery.__init__(self, *args)

    def compose(self, args, optargs):
        if len(args) != 2:
            return T('r.do(', T(T(*(args[1:]), intsp=', '), args[0],
                                intsp=', '), ')')

        if isinstance(self._args[1], Datum):
            args[1] = T('r.expr(', args[1], ')')

        return T(args[1], '.do(', args[0], ')')


class Table(RqlQuery):
    term_type = P_TERM.TABLE
    statement = 'table'

    def insert(self, *args, **kwargs):
        return Insert(self, *[expr(arg) for arg in args], **kwargs)

    def get(self, *args):
        return Get(self, *args)

    def get_all(self, *args, **kwargs):
        return GetAll(self, *args, **kwargs)

    def set_write_hook(self, *args, **kwargs):
        return SetWriteHook(self, *args, **kwargs)

    def get_write_hook(self, *args, **kwargs):
        return GetWriteHook(self, *args, **kwargs)

    def index_create(self, *args, **kwargs):
        if len(args) > 1:
            args = [args[0]] + [func_wrap(arg) for arg in args[1:]]
        return IndexCreate(self, *args, **kwargs)

    def index_drop(self, *args):
        return IndexDrop(self, *args)

    def index_rename(self, *args, **kwargs):
        return IndexRename(self, *args, **kwargs)

    def index_list(self, *args):
        return IndexList(self, *args)

    def index_status(self, *args):
        return IndexStatus(self, *args)

    def index_wait(self, *args):
        return IndexWait(self, *args)

    def status(self, *args):
        return Status(self, *args)

    def config(self, *args):
        return Config(self, *args)

    def wait(self, *args, **kwargs):
        return Wait(self, *args, **kwargs)

    def reconfigure(self, *args, **kwargs):
        return Reconfigure(self, *args, **kwargs)

    def rebalance(self, *args, **kwargs):
        return Rebalance(self, *args, **kwargs)

    def sync(self, *args):
        return Sync(self, *args)

    def grant(self, *args, **kwargs):
        return Grant(self, *args, **kwargs)

    def get_intersecting(self, *args, **kwargs):
        return GetIntersecting(self, *args, **kwargs)

    def get_nearest(self, *args, **kwargs):
        return GetNearest(self, *args, **kwargs)

    def uuid(self, *args, **kwargs):
        return UUID(self, *args, **kwargs)

    def compose(self, args, optargs):
        args.extend([T(k, '=', v) for k, v in dict_items(optargs)])
        if isinstance(self._args[0], DB):
            return T(args[0], '.table(', T(*(args[1:]), intsp=', '), ')')
        else:
            return T('r.table(', T(*(args), intsp=', '), ')')


class Get(RqlMethodQuery):
    term_type = P_TERM.GET
    statement = 'get'


class GetAll(RqlMethodQuery):
    term_type = P_TERM.GET_ALL
    statement = 'get_all'


class GetIntersecting(RqlMethodQuery):
    term_type = P_TERM.GET_INTERSECTING
    statement = 'get_intersecting'


class GetNearest(RqlMethodQuery):
    term_type = P_TERM.GET_NEAREST
    statement = 'get_nearest'


class UUID(RqlMethodQuery):
    term_type = P_TERM.UUID
    statement = 'uuid'


class Reduce(RqlMethodQuery):
    term_type = P_TERM.REDUCE
    statement = 'reduce'


class Sum(RqlMethodQuery):
    term_type = P_TERM.SUM
    statement = 'sum'


class Avg(RqlMethodQuery):
    term_type = P_TERM.AVG
    statement = 'avg'


class Min(RqlMethodQuery):
    term_type = P_TERM.MIN
    statement = 'min'


class Max(RqlMethodQuery):
    term_type = P_TERM.MAX
    statement = 'max'


class Map(RqlMethodQuery):
    term_type = P_TERM.MAP
    statement = 'map'


class Fold(RqlMethodQuery):
    term_type = P_TERM.FOLD
    statement = 'fold'


class Filter(RqlMethodQuery):
    term_type = P_TERM.FILTER
    statement = 'filter'


class ConcatMap(RqlMethodQuery):
    term_type = P_TERM.CONCAT_MAP
    statement = 'concat_map'


class OrderBy(RqlMethodQuery):
    term_type = P_TERM.ORDER_BY
    statement = 'order_by'


class Distinct(RqlMethodQuery):
    term_type = P_TERM.DISTINCT
    statement = 'distinct'


class Count(RqlMethodQuery):
    term_type = P_TERM.COUNT
    statement = 'count'


class Union(RqlMethodQuery):
    term_type = P_TERM.UNION
    statement = 'union'


class Nth(RqlBracketQuery):
    term_type = P_TERM.NTH
    statement = 'nth'


class Match(RqlMethodQuery):
    term_type = P_TERM.MATCH
    statement = 'match'


class ToJsonString(RqlMethodQuery):
    term_type = P_TERM.TO_JSON_STRING
    statement = 'to_json_string'


class Split(RqlMethodQuery):
    term_type = P_TERM.SPLIT
    statement = 'split'


class Upcase(RqlMethodQuery):
    term_type = P_TERM.UPCASE
    statement = 'upcase'


class Downcase(RqlMethodQuery):
    term_type = P_TERM.DOWNCASE
    statement = 'downcase'


class OffsetsOf(RqlMethodQuery):
    term_type = P_TERM.OFFSETS_OF
    statement = 'offsets_of'


class IsEmpty(RqlMethodQuery):
    term_type = P_TERM.IS_EMPTY
    statement = 'is_empty'


class Group(RqlMethodQuery):
    term_type = P_TERM.GROUP
    statement = 'group'


class InnerJoin(RqlMethodQuery):
    term_type = P_TERM.INNER_JOIN
    statement = 'inner_join'


class OuterJoin(RqlMethodQuery):
    term_type = P_TERM.OUTER_JOIN
    statement = 'outer_join'


class EqJoin(RqlMethodQuery):
    term_type = P_TERM.EQ_JOIN
    statement = 'eq_join'


class Zip(RqlMethodQuery):
    term_type = P_TERM.ZIP
    statement = 'zip'


class CoerceTo(RqlMethodQuery):
    term_type = P_TERM.COERCE_TO
    statement = 'coerce_to'


class Ungroup(RqlMethodQuery):
    term_type = P_TERM.UNGROUP
    statement = 'ungroup'


class TypeOf(RqlMethodQuery):
    term_type = P_TERM.TYPE_OF
    statement = 'type_of'


class Update(RqlMethodQuery):
    term_type = P_TERM.UPDATE
    statement = 'update'


class Delete(RqlMethodQuery):
    term_type = P_TERM.DELETE
    statement = 'delete'


class Replace(RqlMethodQuery):
    term_type = P_TERM.REPLACE
    statement = 'replace'


class Insert(RqlMethodQuery):
    term_type = P_TERM.INSERT
    statement = 'insert'


class DbCreate(RqlTopLevelQuery):
    term_type = P_TERM.DB_CREATE
    statement = "db_create"


class DbDrop(RqlTopLevelQuery):
    term_type = P_TERM.DB_DROP
    statement = "db_drop"


class DbList(RqlTopLevelQuery):
    term_type = P_TERM.DB_LIST
    statement = "db_list"


class TableCreate(RqlMethodQuery):
    term_type = P_TERM.TABLE_CREATE
    statement = "table_create"


class TableCreateTL(RqlTopLevelQuery):
    term_type = P_TERM.TABLE_CREATE
    statement = "table_create"


class TableDrop(RqlMethodQuery):
    term_type = P_TERM.TABLE_DROP
    statement = "table_drop"


class TableDropTL(RqlTopLevelQuery):
    term_type = P_TERM.TABLE_DROP
    statement = "table_drop"


class TableList(RqlMethodQuery):
    term_type = P_TERM.TABLE_LIST
    statement = "table_list"


class TableListTL(RqlTopLevelQuery):
    term_type = P_TERM.TABLE_LIST
    statement = "table_list"


class SetWriteHook(RqlMethodQuery):
    term_type = P_TERM.SET_WRITE_HOOK
    statement = 'set_write_hook'


class GetWriteHook(RqlMethodQuery):
    term_type = P_TERM.GET_WRITE_HOOK
    statement = 'get_write_hook'


class IndexCreate(RqlMethodQuery):
    term_type = P_TERM.INDEX_CREATE
    statement = 'index_create'


class IndexDrop(RqlMethodQuery):
    term_type = P_TERM.INDEX_DROP
    statement = 'index_drop'


class IndexRename(RqlMethodQuery):
    term_type = P_TERM.INDEX_RENAME
    statement = 'index_rename'


class IndexList(RqlMethodQuery):
    term_type = P_TERM.INDEX_LIST
    statement = 'index_list'


class IndexStatus(RqlMethodQuery):
    term_type = P_TERM.INDEX_STATUS
    statement = 'index_status'


class IndexWait(RqlMethodQuery):
    term_type = P_TERM.INDEX_WAIT
    statement = 'index_wait'


class Config(RqlMethodQuery):
    term_type = P_TERM.CONFIG
    statement = "config"


class Status(RqlMethodQuery):
    term_type = P_TERM.STATUS
    statement = "status"


class Wait(RqlMethodQuery):
    term_type = P_TERM.WAIT
    statement = "wait"


class Reconfigure(RqlMethodQuery):
    term_type = P_TERM.RECONFIGURE
    statement = 'reconfigure'


class Rebalance(RqlMethodQuery):
    term_type = P_TERM.REBALANCE
    statement = 'rebalance'


class Sync(RqlMethodQuery):
    term_type = P_TERM.SYNC
    statement = 'sync'


class Grant(RqlMethodQuery):
    term_type = P_TERM.GRANT
    statement = 'grant'


class GrantTL(RqlTopLevelQuery):
    term_type = P_TERM.GRANT
    statement = 'grant'


class Branch(RqlTopLevelQuery):
    term_type = P_TERM.BRANCH
    statement = "branch"


class Or(RqlBoolOperQuery):
    term_type = P_TERM.OR
    statement = "or_"
    st_infix = "|"


class And(RqlBoolOperQuery):
    term_type = P_TERM.AND
    statement = "and_"
    st_infix = "&"


class ForEach(RqlMethodQuery):
    term_type = P_TERM.FOR_EACH
    statement = 'for_each'


class Info(RqlMethodQuery):
    term_type = P_TERM.INFO
    statement = 'info'


class InsertAt(RqlMethodQuery):
    term_type = P_TERM.INSERT_AT
    statement = 'insert_at'


class SpliceAt(RqlMethodQuery):
    term_type = P_TERM.SPLICE_AT
    statement = 'splice_at'


class DeleteAt(RqlMethodQuery):
    term_type = P_TERM.DELETE_AT
    statement = 'delete_at'


class ChangeAt(RqlMethodQuery):
    term_type = P_TERM.CHANGE_AT
    statement = 'change_at'


class Sample(RqlMethodQuery):
    term_type = P_TERM.SAMPLE
    statement = 'sample'


class Json(RqlTopLevelQuery):
    term_type = P_TERM.JSON
    statement = 'json'


class Args(RqlTopLevelQuery):
    term_type = P_TERM.ARGS
    statement = 'args'


# Use this class as a wrapper to 'bytes' so we can tell the difference
# in Python2 (when reusing the result of a previous query).
class RqlBinary(bytes):
    def __new__(cls, *args, **kwargs):
        return bytes.__new__(cls, *args, **kwargs)

    def __repr__(self):
        excerpt = binascii.hexlify(self[0:6]).decode('utf-8')
        excerpt = ' '.join([excerpt[i:i + 2]
                            for i in xrange(0, len(excerpt), 2)])
        excerpt = ', \'%s%s\'' % (excerpt, '...' if len(self) > 6 else '') \
                  if len(self) > 0 else ''
        return "<binary, %d byte%s%s>" % (len(self), 's'
                                          if len(self) != 1 else '', excerpt)


class Binary(RqlTopLevelQuery):
    # Note: this term isn't actually serialized, it should exist only
    # in the client
    term_type = P_TERM.BINARY
    statement = 'binary'

    def __init__(self, data):
        # We only allow 'bytes' objects to be serialized as binary
        # Python 2 - `bytes` is equivalent to `str`, either will be accepted
        # Python 3 - `unicode` is equivalent to `str`, neither will be accepted
        if isinstance(data, RqlQuery):
            RqlTopLevelQuery.__init__(self, data)
        elif isinstance(data, unicode):
            raise ReqlDriverCompileError("Cannot convert a unicode string to binary, "
                                         "use `unicode.encode()` to specify the "
                                         "encoding.")
        elif not isinstance(data, bytes):
            raise ReqlDriverCompileError(("Cannot convert %s to binary, convert the "
                                          "object to a `bytes` object first.")
                                         % type(data).__name__)
        else:
            self.base64_data = base64.b64encode(data)

            # Kind of a hack to get around composing
            self._args = []
            self.optargs = {}

    def compose(self, args, optargs):
        if len(self._args) == 0:
            return T('r.', self.statement, '(bytes(<data>))')
        else:
            return RqlTopLevelQuery.compose(self, args, optargs)

    def build(self):
        if len(self._args) == 0:
            return {'$reql_type$': 'BINARY',
                    'data': self.base64_data.decode('utf-8')}
        else:
            return RqlTopLevelQuery.build(self)


class Range(RqlTopLevelQuery):
    term_type = P_TERM.RANGE
    statement = 'range'


class ToISO8601(RqlMethodQuery):
    term_type = P_TERM.TO_ISO8601
    statement = 'to_iso8601'


class During(RqlMethodQuery):
    term_type = P_TERM.DURING
    statement = 'during'


class Date(RqlMethodQuery):
    term_type = P_TERM.DATE
    statement = 'date'


class TimeOfDay(RqlMethodQuery):
    term_type = P_TERM.TIME_OF_DAY
    statement = 'time_of_day'


class Timezone(RqlMethodQuery):
    term_type = P_TERM.TIMEZONE
    statement = 'timezone'


class Year(RqlMethodQuery):
    term_type = P_TERM.YEAR
    statement = 'year'


class Month(RqlMethodQuery):
    term_type = P_TERM.MONTH
    statement = 'month'


class Day(RqlMethodQuery):
    term_type = P_TERM.DAY
    statement = 'day'


class DayOfWeek(RqlMethodQuery):
    term_type = P_TERM.DAY_OF_WEEK
    statement = 'day_of_week'


class DayOfYear(RqlMethodQuery):
    term_type = P_TERM.DAY_OF_YEAR
    statement = 'day_of_year'


class Hours(RqlMethodQuery):
    term_type = P_TERM.HOURS
    statement = 'hours'


class Minutes(RqlMethodQuery):
    term_type = P_TERM.MINUTES
    statement = 'minutes'


class Seconds(RqlMethodQuery):
    term_type = P_TERM.SECONDS
    statement = 'seconds'


class Time(RqlTopLevelQuery):
    term_type = P_TERM.TIME
    statement = 'time'


class ISO8601(RqlTopLevelQuery):
    term_type = P_TERM.ISO8601
    statement = 'iso8601'


class EpochTime(RqlTopLevelQuery):
    term_type = P_TERM.EPOCH_TIME
    statement = 'epoch_time'


class Now(RqlTopLevelQuery):
    term_type = P_TERM.NOW
    statement = 'now'


class InTimezone(RqlMethodQuery):
    term_type = P_TERM.IN_TIMEZONE
    statement = 'in_timezone'


class ToEpochTime(RqlMethodQuery):
    term_type = P_TERM.TO_EPOCH_TIME
    statement = 'to_epoch_time'


class GeoJson(RqlTopLevelQuery):
    term_type = P_TERM.GEOJSON
    statement = 'geojson'


class ToGeoJson(RqlMethodQuery):
    term_type = P_TERM.TO_GEOJSON
    statement = 'to_geojson'


class Point(RqlTopLevelQuery):
    term_type = P_TERM.POINT
    statement = 'point'


class Line(RqlTopLevelQuery):
    term_type = P_TERM.LINE
    statement = 'line'


class Polygon(RqlTopLevelQuery):
    term_type = P_TERM.POLYGON
    statement = 'polygon'


class Distance(RqlMethodQuery):
    term_type = P_TERM.DISTANCE
    statement = 'distance'


class Intersects(RqlMethodQuery):
    term_type = P_TERM.INTERSECTS
    statement = 'intersects'


class Includes(RqlMethodQuery):
    term_type = P_TERM.INCLUDES
    statement = 'includes'


class Circle(RqlTopLevelQuery):
    term_type = P_TERM.CIRCLE
    statement = 'circle'


class Fill(RqlMethodQuery):
    term_type = P_TERM.FILL
    statement = 'fill'


class PolygonSub(RqlMethodQuery):
    term_type = P_TERM.POLYGON_SUB
    statement = 'polygon_sub'


# Returns True if IMPLICIT_VAR is found in the subquery
def _ivar_scan(query):
    if not isinstance(query, RqlQuery):
        return False
    if isinstance(query, ImplicitVar):
        return True
    if any([_ivar_scan(arg) for arg in query._args]):
        return True
    if any([_ivar_scan(arg) for k, arg in dict_items(query.optargs)]):
        return True
    return False


# Called on arguments that should be functions
def func_wrap(val):
    val = expr(val)
    if _ivar_scan(val):
        return Func(lambda x: val)
    return val


class Func(RqlQuery):
    term_type = P_TERM.FUNC
    lock = threading.Lock()
    nextVarId = 1

    def __init__(self, lmbd):
        super(Func, self).__init__()
        vrs = []
        vrids = []
        try:
            code = lmbd.func_code
        except AttributeError:
            code = lmbd.__code__
        for i in xrange(code.co_argcount):
            Func.lock.acquire()
            var_id = Func.nextVarId
            Func.nextVarId += 1
            Func.lock.release()
            vrs.append(Var(var_id))
            vrids.append(var_id)

        self.vrs = vrs
        self._args.extend([MakeArray(*vrids), expr(lmbd(*vrs))])

    def compose(self, args, optargs):
        return T('lambda ', T(*[v.compose([v._args[0].compose(None, None)],
                                          []) for v in self.vrs],
                              intsp=', '), ': ', args[1])


class Asc(RqlTopLevelQuery):
    term_type = P_TERM.ASC
    statement = 'asc'


class Desc(RqlTopLevelQuery):
    term_type = P_TERM.DESC
    statement = 'desc'


class Literal(RqlTopLevelQuery):
    term_type = P_TERM.LITERAL
    statement = 'literal'



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

import socket
import struct

from rethinkdb import ql2_pb2
from rethinkdb.errors import ReqlAuthError, ReqlCursorEmpty, ReqlDriverError, ReqlTimeoutError
from rethinkdb.net import Connection as ConnectionBase, Cursor, Query, Response, maybe_profile
from tornado import gen, iostream
from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado.tcpclient import TCPClient

__all__ = ['Connection']

pResponse = ql2_pb2.Response.ResponseType
pQuery = ql2_pb2.Query.QueryType


@gen.coroutine
def with_absolute_timeout(deadline, generator, **kwargs):
    if deadline is None:
        res = yield generator
    else:
        try:
            res = yield gen.with_timeout(deadline, generator, **kwargs)
        except gen.TimeoutError:
            raise ReqlTimeoutError()
    raise gen.Return(res)


# The Tornado implementation of the Cursor object:
# The `new_response` Future notifies any waiting coroutines that the can attempt
# to grab the next result.  In addition, the waiting coroutine will schedule a
# timeout at the given deadline (if provided), at which point the future will be
# errored.
class TornadoCursor(Cursor):
    def __init__(self, *args, **kwargs):
        Cursor.__init__(self, *args, **kwargs)
        self.new_response = Future()

    def _extend(self, res_buf):
        Cursor._extend(self, res_buf)
        self.new_response.set_result(True)
        self.new_response = Future()

    # Convenience function so users know when they've hit the end of the cursor
    # without having to catch an exception
    @gen.coroutine
    def fetch_next(self, wait=True):
        timeout = Cursor._wait_to_timeout(wait)
        deadline = None if timeout is None else self.conn._io_loop.time() + timeout
        while len(self.items) == 0 and self.error is None:
            self._maybe_fetch_batch()
            yield with_absolute_timeout(deadline, self.new_response)
        # If there is a (non-empty) error to be received, we return True, so the
        # user will receive it on the next `next` call.
        raise gen.Return(len(self.items) != 0 or not isinstance(self.error, ReqlCursorEmpty))

    def _empty_error(self):
        # We do not have ReqlCursorEmpty inherit from StopIteration as that interferes
        # with Tornado's gen.coroutine and is the equivalent of gen.Return(None).
        return ReqlCursorEmpty()

    @gen.coroutine
    def _get_next(self, timeout):
        deadline = None if timeout is None else self.conn._io_loop.time() + timeout
        while len(self.items) == 0:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            yield with_absolute_timeout(deadline, self.new_response)
        raise gen.Return(self.items.popleft())


class ConnectionInstance(object):
    def __init__(self, parent, io_loop=None):
        self._parent = parent
        self._closing = False
        self._user_queries = {}
        self._cursor_cache = {}
        self._ready = Future()
        self._io_loop = io_loop
        self._stream = None
        if self._io_loop is None:
            self._io_loop = IOLoop.current()
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def client_port(self):
        if self.is_open():
            return self._socket.getsockname()[1]

    def client_address(self):
        if self.is_open():
            return self._socket.getsockname()[0]

    @gen.coroutine
    def connect(self, timeout):
        deadline = None if timeout is None else self._io_loop.time() + timeout

        try:
            if len(self._parent.ssl) > 0:
                ssl_options = {}
                if self._parent.ssl["ca_certs"]:
                    ssl_options['ca_certs'] = self._parent.ssl["ca_certs"]
                    ssl_options['cert_reqs'] = 2  # ssl.CERT_REQUIRED
                stream_future = TCPClient().connect(self._parent.host, self._parent.port,
                                                    ssl_options=ssl_options)
            else:
                stream_future = TCPClient().connect(self._parent.host, self._parent.port)

            self._stream = yield with_absolute_timeout(
                deadline,
                stream_future,
                quiet_exceptions=(iostream.StreamClosedError))
        except Exception as err:
            raise ReqlDriverError('Could not connect to %s:%s. Error: %s' %
                                  (self._parent.host, self._parent.port, str(err)))

        self._stream.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._stream.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        try:
            self._parent.handshake.reset()
            response = None
            while True:
                request = self._parent.handshake.next_message(response)
                if request is None:
                    break
                # This may happen in the `V1_0` protocol where we send two requests as
                # an optimization, then need to read each separately
                if request is not "":
                    self._stream.write(request)

                response = yield with_absolute_timeout(
                    deadline,
                    self._stream.read_until(b'\0'),
                    quiet_exceptions=(iostream.StreamClosedError))
                response = response[:-1]
        except ReqlAuthError:
            try:
                self._stream.close()
            except iostream.StreamClosedError:
                pass
            raise
        except ReqlTimeoutError:
            try:
                self._stream.close()
            except iostream.StreamClosedError:
                pass
            raise ReqlTimeoutError(self._parent.host, self._parent.port)
        except Exception as err:
            try:
                self._stream.close()
            except iostream.StreamClosedError:
                pass
            raise ReqlDriverError(
                'Connection interrupted during handshake with %s:%s. Error: %s' %
                (self._parent.host, self._parent.port, str(err)))

        # Start a parallel function to perform reads
        self._io_loop.add_callback(self._reader)
        raise gen.Return(self._parent)

    def is_open(self):
        return not self._stream.closed()

    @gen.coroutine
    def close(self, noreply_wait=False, token=None, exception=None):
        self._closing = True
        if exception is not None:
            err_message = "Connection is closed (%s)." % str(exception)
        else:
            err_message = "Connection is closed."

        # Cursors may remove themselves when errored, so copy a list of them
        for cursor in list(self._cursor_cache.values()):
            cursor._error(err_message)

        for query, future in iter(self._user_queries.values()):
            future.set_exception(ReqlDriverError(err_message))

        self._user_queries = {}
        self._cursor_cache = {}

        if noreply_wait:
            noreply = Query(pQuery.NOREPLY_WAIT, token, None, None)
            yield self.run_query(noreply, False)

        try:
            self._stream.close()
        except iostream.StreamClosedError:
            pass
        raise gen.Return(None)

    @gen.coroutine
    def run_query(self, query, noreply):
        yield self._stream.write(query.serialize(self._parent._get_json_encoder(query)))
        if noreply:
            raise gen.Return(None)

        response_future = Future()
        self._user_queries[query.token] = (query, response_future)
        res = yield response_future
        raise gen.Return(res)

    # The _reader coroutine runs in its own context at the top level of the
    # Tornado.IOLoop it was created with.  It runs in parallel, reading responses
    # off of the socket and forwarding them to the appropriate Future or Cursor.
    # This is shut down as a consequence of closing the stream, or an error in the
    # socket/protocol from the server.  Unexpected errors in this coroutine will
    # close the ConnectionInstance and be passed to any open Futures or Cursors.
    @gen.coroutine
    def _reader(self):
        try:
            while True:
                buf = yield self._stream.read_bytes(12)
                (token, length,) = struct.unpack("<qL", buf)
                buf = yield self._stream.read_bytes(length)

                cursor = self._cursor_cache.get(token)
                if cursor is not None:
                    cursor._extend(buf)
                elif token in self._user_queries:
                    # Do not pop the query from the dict until later, so
                    # we don't lose track of it in case of an exception
                    query, future = self._user_queries[token]
                    res = Response(token, buf,
                                   self._parent._get_json_decoder(query))
                    if res.type == pResponse.SUCCESS_ATOM:
                        future.set_result(maybe_profile(res.data[0], res))
                    elif res.type in (pResponse.SUCCESS_SEQUENCE,
                                      pResponse.SUCCESS_PARTIAL):
                        cursor = TornadoCursor(self, query, res)
                        future.set_result(maybe_profile(cursor, res))
                    elif res.type == pResponse.WAIT_COMPLETE:
                        future.set_result(None)
                    elif res.type == pResponse.SERVER_INFO:
                        future.set_result(res.data[0])
                    else:
                        future.set_exception(res.make_error(query))
                    del self._user_queries[token]
                elif not self._closing:
                    raise ReqlDriverError("Unexpected response received.")
        except Exception as ex:
            if not self._closing:
                yield self.close(exception=ex)


# Wrap functions from the base connection class that may throw - these will
# put any exception inside a Future and return it.
class Connection(ConnectionBase):
    def __init__(self, *args, **kwargs):
        ConnectionBase.__init__(self, ConnectionInstance, *args, **kwargs)

    @gen.coroutine
    def reconnect(self, noreply_wait=True, timeout=None):
        # We close before reconnect so reconnect doesn't try to close us
        # and then fail to return the Future (this is a little awkward).
        yield self.close(noreply_wait)
        res = yield ConnectionBase.reconnect(self, noreply_wait, timeout)
        raise gen.Return(res)

    @gen.coroutine
    def close(self, *args, **kwargs):
        if self._instance is None:
            res = None
        else:
            res = yield ConnectionBase.close(self, *args, **kwargs)
        raise gen.Return(res)

    @gen.coroutine
    def noreply_wait(self):
        res = yield ConnectionBase.noreply_wait(self)
        raise gen.Return(res)

    @gen.coroutine
    def server(self):
        res = yield ConnectionBase.server(self)
        raise gen.Return(res)

    @gen.coroutine
    def _start(self, term, **global_optargs):
        res = yield ConnectionBase._start(self, term, **global_optargs)
        raise gen.Return(res)

    @gen.coroutine
    def _continue(self, cursor):
        res = yield ConnectionBase._continue(self, cursor)
        raise gen.Return(res)

    @gen.coroutine
    def _stop(self, cursor):
        res = yield ConnectionBase._stop(self, cursor)
        raise gen.Return(res)



#!/usr/bin/env python

# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

'''`rethinkdb import` loads data into a RethinkDB cluster'''

from __future__ import print_function

import codecs
import collections
import csv
import ctypes
import json
import multiprocessing
import optparse
import os
import signal
import sys
import time
import traceback
import six
from multiprocessing.queues import Queue, SimpleQueue

from rethinkdb import ast, errors, query, utils_common
from rethinkdb.logger import default_logger

try:
    unicode
except NameError:
    unicode = str

try:
    from Queue import Empty, Full
except ImportError:
    from queue import Empty, Full


# json parameters
JSON_READ_CHUNK_SIZE = 128 * 1024
JSON_MAX_BUFFER_SIZE = 128 * 1024 * 1024
MAX_NESTING_DEPTH = 100

Error = collections.namedtuple("Error", ["message", "traceback", "file"])


class SourceFile(object):
    format = None  # set by subclasses

    name = None

    db = None
    table = None
    primary_key = None
    indexes = None
    write_hook = None
    source_options = None

    start_time = None
    end_time = None

    query_runner = None

    _source = None  # open filehandle for the source

    # - internal synchronization variables

    _bytes_size = None
    _bytes_read = None  # -1 until started

    _total_rows = None  # -1 until known
    _rows_read = None
    _rows_written = None

    def __init__(
            self,
            source,
            db,
            table,
            query_runner,
            primary_key=None,
            indexes=None,
            write_hook=None,
            source_options=None):

        if self.format is None:
            raise AssertionError('{class_name} must have a format'.format(class_name=self.__class__.__name__))

        if self.db == 'rethinkdb':
            raise AssertionError('Can not import tables into the system database')


        # query_runner
        if not isinstance(query_runner, utils_common.RetryQuery):
            raise AssertionError('Query runner is not instance of RetryQuery')

        self.query_runner = query_runner

        # reporting information
        self._bytes_size = multiprocessing.Value(ctypes.c_longlong, -1)
        self._bytes_read = multiprocessing.Value(ctypes.c_longlong, -1)

        self._total_rows = multiprocessing.Value(ctypes.c_longlong, -1)
        self._rows_read = multiprocessing.Value(ctypes.c_longlong, 0)
        self._rows_written = multiprocessing.Value(ctypes.c_longlong, 0)

        # source
        if hasattr(source, 'read'):
            if unicode != str or 'b' in source.mode:
                # Python2.x or binary file, assume utf-8 encoding
                self._source = codecs.getreader("utf-8")(source)
            else:
                # assume that it has the right encoding on it
                self._source = source
        else:
            try:
                self._source = codecs.open(source, mode="r", encoding="utf-8")
            except IOError as exc:
                default_logger.exception(exc)
                raise ValueError('Unable to open source file "%s": %s' % (str(source), str(exc)))

        if hasattr(self._source, 'name') and self._source.name and os.path.isfile(self._source.name):
            self._bytes_size.value = os.path.getsize(source)
            if self._bytes_size.value == 0:
                raise ValueError('Source is zero-length: %s' % source)

        # table info
        self.db = db
        self.table = table
        self.primary_key = primary_key
        self.indexes = indexes or []
        self.write_hook = write_hook or []

        # options
        self.source_options = source_options or {}

        # name
        if hasattr(self._source, 'name') and self._source.name:
            self.name = os.path.basename(self._source.name)
        else:
            self.name = '%s.%s' % (self.db, self.table)

    def __hash__(self):
        return hash((self.db, self.table))

    def get_line(self):
        '''Returns a single line from the file'''
        raise NotImplementedError('This needs to be implemented on the %s subclass' % self.format)

    # - bytes
    @property
    def bytes_size(self):
        return self._bytes_size.value

    @bytes_size.setter
    def bytes_size(self, value):
        self._bytes_size.value = value

    @property
    def bytes_read(self):
        return self._bytes_read.value

    @bytes_read.setter
    def bytes_read(self, value):
        self._bytes_read.value = value

    # - rows
    @property
    def total_rows(self):
        return self._total_rows.value

    @total_rows.setter
    def total_rows(self, value):
        self._total_rows.value = value

    @property
    def rows_read(self):
        return self._rows_read.value

    @rows_read.setter
    def rows_read(self, value):
        self._rows_read.value = value

    @property
    def rows_written(self):
        return self._rows_written.value

    def add_rows_written(self, increment):  # we have multiple writers to coordinate
        with self._rows_written.get_lock():
            self._rows_written.value += increment

    # - percent done
    @property
    def percent_done(self):
        '''return a float between 0 and 1 for a reasonable guess of percentage complete'''
        # assume that reading takes 50% of the time and writing the other 50%
        completed = 0.0  # of 2.0

        # - add read percentage
        if self._bytes_size.value <= 0 or self._bytes_size.value <= self._bytes_read.value:
            completed += 1.0
        elif self._bytes_read.value < 0 and self._total_rows.value >= 0:
            # done by rows read
            if self._rows_read > 0:
                completed += float(self._rows_read) / float(self._total_rows.value)
        else:
            # done by bytes read
            if self._bytes_read.value > 0:
                completed += float(self._bytes_read.value) / float(self._bytes_size.value)

        # - add written percentage
        if self._rows_read.value or self._rows_written.value:
            total_rows = float(self._total_rows.value)
            if total_rows == 0:
                completed += 1.0
            elif total_rows < 0:
                # a guesstimate
                per_row_size = float(self._bytes_read.value) / float(self._rows_read.value)
                total_rows = float(self._rows_read.value) + \
                    (float(self._bytes_size.value - self._bytes_read.value) / per_row_size)
                completed += float(self._rows_written.value) / total_rows
            else:
                # accurate count
                completed += float(self._rows_written.value) / total_rows

        # - return the value
        return completed * 0.5

    def setup_table(self):
        '''Ensure that the db, table, and indexes exist and are correct'''

        # - ensure the table exists and is ready
        self.query_runner(
            "create table: %s.%s" % (self.db, self.table),
            ast.expr([self.table]).set_difference(
                query.db(self.db).table_list()
            ).for_each(query.db(self.db).table_create(
                query.row, **self.source_options.create_args if 'create_args' in self.source_options else {})
            )
        )

        self.query_runner(
            "wait for %s.%s" % (self.db, self.table),
            query.db(self.db).table(self.table).wait(timeout=30)
        )

        # - ensure that the primary key on the table is correct
        primary_key = self.query_runner(
            "primary key %s.%s" % (self.db, self.table),
            query.db(self.db).table(self.table).info()["primary_key"],
        )
        if self.primary_key is None:
            self.primary_key = primary_key
        elif primary_key != self.primary_key:
            raise RuntimeError(
                "Error: table %s.%s primary key was `%s` rather than the expected: %s" %
                (self.db, self.table, primary_key, self.primary_key))

    def restore_indexes(self, warning_queue):
        # recreate secondary indexes - dropping existing on the assumption they are wrong
        if self.indexes:
            existing_indexes = self.query_runner(
                "indexes from: %s.%s" %
                (self.db, self.table), query.db(
                    self.db).table(
                    self.table).index_list())
            try:
                created_indexes = []
                for index in self.indexes:
                    if index["index"] in existing_indexes:  # drop existing versions
                        self.query_runner(
                            "drop index: %s.%s:%s" % (self.db, self.table, index["index"]),
                            query.db(self.db).table(self.table).index_drop(index["index"])
                        )
                    self.query_runner(
                        "create index: %s.%s:%s" % (self.db, self.table, index["index"]),
                        query.db(self.db).table(self.table).index_create(index["index"], index["function"])
                    )
                    created_indexes.append(index["index"])

                # wait for all of the created indexes to build
                self.query_runner(
                    "waiting for indexes on %s.%s" % (self.db, self.table),
                    query.db(self.db).table(self.table).index_wait(query.args(created_indexes))
                )
            except RuntimeError:
                exception_type, exception_class, trcback = sys.exc_info()
                warning_queue.put((exception_type, exception_class, traceback.extract_tb(trcback), self._source.name))

        if self.write_hook:
            self.query_runner(
                "Write hook from: %s.%s" %
                (self.db, self.table), query.db(
                    self.db).table(
                    self.table).get_write_hook())
            try:
                self.query_runner(
                    "drop hook: %s.%s" % (self.db, self.table),
                    query.db(self.db).table(self.table).set_write_hook(None)
                )
                self.query_runner(
                    "create hook: %s.%s:%s" % (self.db, self.table, self.write_hook),
                    query.db(self.db).table(self.table).set_write_hook(self.write_hook["function"])
                )
            except RuntimeError:
                exception_type, exception_class, trcback = sys.exc_info()
                warning_queue.put((exception_type, exception_class, traceback.extract_tb(trcback), self._source.name))

    def batches(self, batch_size=None, warning_queue=None):

        # setup table
        self.setup_table()

        # default batch_size
        if batch_size is None:
            batch_size = utils_common.default_batch_size
        else:
            batch_size = int(batch_size)

        if batch_size <= 0:
            raise AssertionError('Batch size can not be less than one')

        # setup
        self.setup_file(warning_queue=warning_queue)

        # - yield batches

        batch = []
        try:
            need_more_data = False
            while True:
                if need_more_data:
                    self.fill_buffer()
                    need_more_data = False

                while len(batch) < batch_size:
                    try:
                        row = self.get_line()
                        # ToDo: validate the line
                        batch.append(row)
                    except NeedMoreData:
                        need_more_data = True
                        break
                else:
                    yield batch
                    batch = []

        except StopIteration as e:
            # yield any final batch
            if batch:
                yield batch

            # - check the end of the file

            self.teardown()

            # - rebuild indexes
            if self.indexes:
                self.restore_indexes(warning_queue)

    def setup_file(self, warning_queue=None):
        raise NotImplementedError("Subclasses need to implement this")

    def teardown(self):
        pass

    def read_to_queue(
            self,
            work_queue,
            exit_event,
            error_queue,
            warning_queue,
            timing_queue,
            fields=None,
            ignore_signals=True,
            batch_size=None):
        if ignore_signals:  # ToDo: work out when we are in a worker process automatically
            signal.signal(signal.SIGINT, signal.SIG_IGN)  # workers should ignore these

        if batch_size is None:
            batch_size = utils_common.default_batch_size

        self.start_time = time.time()
        try:
            timePoint = time.time()
            for batch in self.batches(warning_queue=warning_queue):
                timing_queue.put(('reader_work', time.time() - timePoint))
                timePoint = time.time()

                # apply the fields filter
                if fields:
                    for row in batch:
                        for key in [x for x in row.keys() if x not in fields]:
                            del row[key]

                while not exit_event.is_set():
                    try:
                        work_queue.put((self.db, self.table, batch), timeout=0.1)
                        self._rows_read.value += len(batch)
                        break
                    except Full:
                        pass
                else:
                    break
                timing_queue.put(('reader_wait', time.time() - timePoint))
                timePoint = time.time()

        # - report relevant errors
        except Exception as exc:
            default_logger.exception(exc)
            error_queue.put(Error(str(exc), traceback.format_exc(), self.name))
            exit_event.set()
            raise
        finally:
            self.end_time = time.time()


class NeedMoreData(Exception):
    pass


class JsonSourceFile(SourceFile):
    format = 'json'

    decoder = json.JSONDecoder()
    json_array = None
    found_first = False

    _buffer_size = JSON_READ_CHUNK_SIZE
    _buffer_str = None
    _buffer_pos = None
    _buffer_end = None

    def fill_buffer(self):
        if self._buffer_str is None:
            self._buffer_str = ''
            self._buffer_pos = 0
            self._buffer_end = 0
        elif self._buffer_pos == 0:
            # double the buffer under the assumption that the documents are too large to fit
            if self._buffer_size == JSON_MAX_BUFFER_SIZE:
                raise Exception(
                    "Error: JSON max buffer size exceeded on file %s (from position %d). Use '--max-document-size' to "
                    "extend your buffer." %
                    (self.name, self.bytes_processed))
            self._buffer_size = min(self._buffer_size * 2, JSON_MAX_BUFFER_SIZE)

        # add more data
        read_target = self._buffer_size - self._buffer_end + self._buffer_pos

        if read_target < 1:
            raise AssertionError('Can not set the read target and full the buffer')

        new_chunk = self._source.read(read_target)

        if len(new_chunk) == 0:
            raise StopIteration()  # file ended

        self._buffer_str = self._buffer_str[self._buffer_pos:] + new_chunk
        self._bytes_read.value += len(new_chunk)

        # reset markers
        self._buffer_pos = 0
        self._buffer_end = len(self._buffer_str) - 1

    def get_line(self):
        '''Return a line from the current _buffer_str, or raise NeedMoreData trying'''

        # advance over any whitespace
        self._buffer_pos = json.decoder.WHITESPACE.match(self._buffer_str, self._buffer_pos).end()
        if self._buffer_pos >= self._buffer_end:
            raise NeedMoreData()

        # read over a comma if we are not the first item in a json_array
        if self.json_array and self.found_first and self._buffer_str[self._buffer_pos] == ",":
            self._buffer_pos += 1
            if self._buffer_pos >= self._buffer_end:
                raise NeedMoreData()

        # advance over any post-comma whitespace
        self._buffer_pos = json.decoder.WHITESPACE.match(self._buffer_str, self._buffer_pos).end()
        if self._buffer_pos >= self._buffer_end:
            raise NeedMoreData()

        # parse and return an object
        try:
            row, self._buffer_pos = self.decoder.raw_decode(self._buffer_str, idx=self._buffer_pos)
            self.found_first = True
            return row
        except (ValueError, IndexError):
            raise NeedMoreData()

    def setup_file(self, warning_queue=None):
        # - move to the first record

        # advance through any leading whitespace
        while True:
            self.fill_buffer()
            self._buffer_pos = json.decoder.WHITESPACE.match(self._buffer_str, 0).end()
            if self._buffer_pos == 0:
                break

        # check the first character
        try:
            if self._buffer_str[0] == "[":
                self.json_array = True
                self._buffer_pos = 1
            elif self._buffer_str[0] == "{":
                self.json_array = False
            else:
                raise ValueError("Error: JSON format not recognized - file does not begin with an object or array")
        except IndexError:
            raise ValueError("Error: JSON file was empty of content")

    def teardown(self):

        # - check the end of the file
        # note: fill_buffer should have guaranteed that we have only the data in the end

        # advance through any leading whitespace
        self._buffer_pos = json.decoder.WHITESPACE.match(self._buffer_str, self._buffer_pos).end()

        # check the end of the array if we have it
        if self.json_array:
            if self._buffer_str[self._buffer_pos] != "]":
                snippit = self._buffer_str[self._buffer_pos:]
                extra = '' if len(snippit) <= 100 else ' and %d more characters' % (len(snippit) - 100)
                raise ValueError("Error: JSON array did not end cleanly, rather with: <<%s>>%s" %
                                 (snippit[:100], extra))
            self._buffer_pos += 1

        # advance through any trailing whitespace
        self._buffer_pos = json.decoder.WHITESPACE.match(self._buffer_str, self._buffer_pos).end()
        snippit = self._buffer_str[self._buffer_pos:]
        if len(snippit) > 0:
            extra = '' if len(snippit) <= 100 else ' and %d more characters' % (len(snippit) - 100)
            raise ValueError("Error: extra data after JSON data: <<%s>>%s" % (snippit[:100], extra))


class CsvSourceFile(SourceFile):
    format = "csv"

    no_header_row = False
    custom_header = None

    _reader = None  # instance of csv.reader
    _columns = None  # name of the columns

    def __init__(self, *args, **kwargs):
        if 'source_options' in kwargs and isinstance(kwargs['source_options'], dict):
            if 'no_header_row' in kwargs['source_options']:
                self.no_header_row = kwargs['source_options']['no_header_row']
            if 'custom_header' in kwargs['source_options']:
                self.custom_header = kwargs['source_options']['custom_header']

        super(CsvSourceFile, self).__init__(*args, **kwargs)

    def byte_counter(self):
        '''Generator for getting a byte count on a file being used'''

        for line in self._source:
            self._bytes_read.value += len(line)
            if unicode != str:
                yield line.encode("utf-8")  # Python2.x csv module does not really handle unicode
            else:
                yield line

    def setup_file(self, warning_queue=None):
        # - setup csv.reader with a byte counter wrapper

        self._reader = csv.reader(self.byte_counter())

        # - get the header information for column names

        if not self.no_header_row:
            self._columns = next(self._reader)

        # field names may override fields from the header
        if self.custom_header is not None:
            if not self.no_header_row:
                warning_queue.put("Ignoring header row on %s: %s" % (self.name, str(self._columns)))
            self._columns = self.custom_header
        elif self.no_header_row:
            raise ValueError("Error: No field name information available")

    def get_line(self):
        raw_row = next(self._reader)
        if len(self._columns) != len(raw_row):
            raise Exception(
                "Error: '%s' line %d has an inconsistent number of columns: %s" %
                (self.name, self._reader.line_num, str(raw_row)))

        row = {}
        for key, value in zip(self._columns, raw_row):  # note: we import all csv fields as strings
            # treat empty fields as no entry rather than empty string
            if value == '':
                continue
            row[key] = value if str == unicode else unicode(value, encoding="utf-8")

        return row

# ==


usage = """rethinkdb import -d DIR [-c HOST:PORT] [--tls-cert FILENAME] [-p] [--password-file FILENAME]
      [--force] [-i (DB | DB.TABLE)] [--clients NUM]
      [--shards NUM_SHARDS] [--replicas NUM_REPLICAS]
  rethinkdb import -f FILE --table DB.TABLE [-c HOST:PORT] [--tls-cert FILENAME] [-p] [--password-file FILENAME]
      [--force] [--clients NUM] [--format (csv | json)] [--pkey PRIMARY_KEY]
      [--shards NUM_SHARDS] [--replicas NUM_REPLICAS]
      [--delimiter CHARACTER] [--custom-header FIELD,FIELD... [--no-header]]"""

help_epilog = '''
EXAMPLES:

rethinkdb import -d rdb_export -c mnemosyne:39500 --clients 128
  Import data into a cluster running on host 'mnemosyne' with a client port at 39500,
  using 128 client connections and the named export directory.

rethinkdb import -f site_history.csv --format csv --table test.history --pkey count
  Import data into a local cluster and the table 'history' in the 'test' database,
  using the named CSV file, and using the 'count' field as the primary key.

rethinkdb import -d rdb_export -c hades -p -i test
  Import data into a cluster running on host 'hades' which requires a password,
  using only the database 'test' from the named export directory.

rethinkdb import -f subscriber_info.json --fields id,name,hashtag --force
  Import data into a local cluster using the named JSON file, and only the fields
  'id', 'name', and 'hashtag', overwriting any existing rows with the same primary key.

rethinkdb import -f user_data.csv --delimiter ';' --no-header --custom-header id,name,number
  Import data into a local cluster using the named CSV file with no header and instead
  use the fields 'id', 'name', and 'number', the delimiter is a semicolon (rather than
  a comma).
'''


def parse_options(argv, prog=None):
    parser = utils_common.CommonOptionsParser(usage=usage, epilog=help_epilog, prog=prog)

    parser.add_option("--clients", dest="clients", metavar="CLIENTS", default=8,
                      help="client connections to use (default: 8)", type="pos_int")
    parser.add_option("--hard-durability", dest="durability", action="store_const", default="soft",
                      help="use hard durability writes (slower, uses less memory)", const="hard")
    parser.add_option("--force", dest="force", action="store_true", default=False,
                      help="import even if a table already exists, overwriting duplicate primary keys")

    parser.add_option(
        "--batch-size",
        dest="batch_size",
        default=utils_common.default_batch_size,
        help=optparse.SUPPRESS_HELP,
        type="pos_int")

    # Replication settings
    replication_options_group = optparse.OptionGroup(parser, "Replication Options")
    replication_options_group.add_option(
        "--shards",
        dest="create_args",
        metavar="SHARDS",
        help="shards to setup on created tables (default: 1)",
        type="pos_int",
        action="add_key")
    replication_options_group.add_option(
        "--replicas",
        dest="create_args",
        metavar="REPLICAS",
        help="replicas to setup on created tables (default: 1)",
        type="pos_int",
        action="add_key")
    parser.add_option_group(replication_options_group)

    # Directory import options
    dir_import_group = optparse.OptionGroup(parser, "Directory Import Options")
    dir_import_group.add_option(
        "-d",
        "--directory",
        dest="directory",
        metavar="DIRECTORY",
        default=None,
        help="directory to import data from")
    dir_import_group.add_option(
        "-i",
        "--import",
        dest="db_tables",
        metavar="DB|DB.TABLE",
        default=[],
        help="restore only the given database or table (may be specified multiple times)",
        action="append",
        type="db_table")
    dir_import_group.add_option(
        "--no-secondary-indexes",
        dest="indexes",
        action="store_false",
        default=None,
        help="do not create secondary indexes")
    parser.add_option_group(dir_import_group)

    # File import options
    file_import_group = optparse.OptionGroup(parser, "File Import Options")
    file_import_group.add_option(
        "-f",
        "--file",
        dest="file",
        metavar="FILE",
        default=None,
        help="file to import data from",
        type="file")
    file_import_group.add_option("--table", dest="import_table", metavar="DB.TABLE",
                               default=None, help="table to import the data into")
    file_import_group.add_option("--fields", dest="fields", metavar="FIELD,...", default=None,
                               help="limit which fields to use when importing one table")
    file_import_group.add_option(
        "--format",
        dest="format",
        metavar="json|csv",
        default=None,
        help="format of the file (default: json, accepts newline delimited json)",
        type="choice",
        choices=[
            "json",
            "csv"])
    file_import_group.add_option(
        "--pkey",
        dest="create_args",
        metavar="PRIMARY_KEY",
        default=None,
        help="field to use as the primary key in the table",
        action="add_key")
    parser.add_option_group(file_import_group)

    # CSV import options
    csv_import_group = optparse.OptionGroup(parser, "CSV Options")
    csv_import_group.add_option(
        "--delimiter",
        dest="delimiter",
        metavar="CHARACTER",
        default=None,
        help="character separating fields, or '\\t' for tab")
    csv_import_group.add_option("--no-header", dest="no_header", action="store_true",
                              default=None, help="do not read in a header of field names")
    csv_import_group.add_option(
        "--custom-header",
        dest="custom_header",
        metavar="FIELD,...",
        default=None,
        help="header to use (overriding file header), must be specified if --no-header")
    parser.add_option_group(csv_import_group)

    # JSON import options
    json_options_group = optparse.OptionGroup(parser, "JSON Options")
    json_options_group.add_option(
        "--max-document-size",
        dest="max_document_size",
        metavar="MAX_SIZE",
        default=0,
        help="maximum allowed size (bytes) for a single JSON document (default: 128MiB)",
        type="pos_int")
    json_options_group.add_option(
        "--max-nesting-depth",
        dest="max_nesting_depth",
        metavar="MAX_DEPTH",
        default=0,
        help="maximum depth of the JSON documents (default: 100)",
        type="pos_int")
    parser.add_option_group(json_options_group)

    options, args = parser.parse_args(argv)

    # Check validity of arguments

    if len(args) != 0:
        raise parser.error("No positional arguments supported. Unrecognized option(s): %s" % args)

    # - create_args
    if options.create_args is None:
        options.create_args = {}

    # - options based on file/directory import

    if options.directory and options.file:
        parser.error("-f/--file and -d/--directory can not be used together")

    elif options.directory:
        if not os.path.exists(options.directory):
            parser.error("-d/--directory does not exist: %s" % options.directory)
        if not os.path.isdir(options.directory):
            parser.error("-d/--directory is not a directory: %s" % options.directory)
        options.directory = os.path.realpath(options.directory)

        # disallow invalid options
        if options.import_table:
            parser.error("--table option is not valid when importing a directory")
        if options.fields:
            parser.error("--fields option is not valid when importing a directory")
        if options.format:
            parser.error("--format option is not valid when importing a directory")
        if options.create_args:
            parser.error("--pkey option is not valid when importing a directory")

        if options.delimiter:
            parser.error("--delimiter option is not valid when importing a directory")
        if options.no_header:
            parser.error("--no-header option is not valid when importing a directory")
        if options.custom_header:
            parser.error("table create options are not valid when importing a directory: %s" %
                         ", ".join([x.lower().replace("_", " ") for x in options.custom_header.keys()]))

        # check valid options
        if not os.path.isdir(options.directory):
            parser.error("Directory to import does not exist: %s" % options.directory)

        if options.fields and (len(options.db_tables) > 1 or options.db_tables[0].table is None):
            parser.error("--fields option can only be used when importing a single table")

    elif options.file:
        if not os.path.exists(options.file):
            parser.error("-f/--file does not exist: %s" % options.file)
        if not os.path.isfile(options.file):
            parser.error("-f/--file is not a file: %s" % options.file)
        options.file = os.path.realpath(options.file)

        # format
        if options.format is None:
            options.format = os.path.splitext(options.file)[1].lstrip('.')

        # import_table
        if options.import_table:
            res = utils_common._tableNameRegex.match(options.import_table)
            if res and res.group("table"):
                options.import_table = utils_common.DbTable(res.group("db"), res.group("table"))
            else:
                parser.error("Invalid --table option: %s" % options.import_table)
        else:
            parser.error("A value is required for --table when importing from a file")

        # fields
        options.fields = options.fields.split(",") if options.fields else None

        # disallow invalid options
        if options.db_tables:
            parser.error("-i/--import can only be used when importing a directory")
        if options.indexes:
            parser.error("--no-secondary-indexes can only be used when importing a directory")

        if options.format == "csv":
            # disallow invalid options
            if options.max_document_size:
                parser.error("--max_document_size only affects importing JSON documents")

            # delimiter
            if options.delimiter is None:
                options.delimiter = ","
            elif options.delimiter == "\\t":
                options.delimiter = "\t"
            elif len(options.delimiter) != 1:
                parser.error("Specify exactly one character for the --delimiter option: %s" % options.delimiter)

            # no_header
            if options.no_header is None:
                options.no_header = False
            elif options.custom_header is None:
                parser.error("--custom-header is required if --no-header is specified")

            # custom_header
            if options.custom_header:
                options.custom_header = options.custom_header.split(",")

        elif options.format == "json":
            # disallow invalid options
            if options.delimiter is not None:
                parser.error("--delimiter option is not valid for json files")
            if options.no_header:
                parser.error("--no-header option is not valid for json files")
            if options.custom_header is not None:
                parser.error("--custom-header option is not valid for json files")

            # default options
            options.format = "json"

            if options.max_document_size > 0:
                global JSON_MAX_BUFFER_SIZE
                JSON_MAX_BUFFER_SIZE = options.max_document_size

            options.file = os.path.abspath(options.file)

        else:
            parser.error("Unrecognized file format: %s" % options.format)

    else:
        parser.error("Either -f/--file or -d/--directory is required")

    # --

    # max_nesting_depth
    if options.max_nesting_depth > 0:
        global MAX_NESTING_DEPTH
        MAX_NESTING_DEPTH = options.max_nesting_depth

    # --

    return options

# This is run for each client requested, and accepts tasks from the reader processes


def table_writer(tables, options, work_queue, error_queue, warning_queue, exit_event, timing_queue):
    signal.signal(signal.SIGINT, signal.SIG_IGN)  # workers should ignore these
    db = table = batch = None

    try:
        conflict_action = "replace" if options.force else "error"
        timePoint = time.time()
        while not exit_event.is_set():
            # get a batch
            try:
                db, table, batch = work_queue.get(timeout=0.1)
            except Empty:
                continue
            timing_queue.put(('writer_wait', time.time() - timePoint))
            timePoint = time.time()

            # shut down when appropriate
            if isinstance(batch, StopIteration):
                return

            # find the table we are working on
            table_info = tables[(db, table)]
            tbl = query.db(db).table(table)

            # write the batch to the database
            try:
                res = options.retryQuery(
                    "write batch to %s.%s" %
                    (db,
                     table),
                    tbl.insert(
                        ast.expr(
                            batch,
                            nesting_depth=MAX_NESTING_DEPTH),
                        durability=options.durability,
                        conflict=conflict_action,
                        ))

                if res["errors"] > 0:
                    raise RuntimeError("Error when importing into table '%s.%s': %s" % (db, table, res["first_error"]))
                modified = res["inserted"] + res["replaced"] + res["unchanged"]
                if modified != len(batch):
                    raise RuntimeError(
                        "The inserted/replaced/unchanged number did not match when importing into table '%s.%s': %s" % (
                            db, table, res["first_error"]
                        )
                    )

                table_info.add_rows_written(modified)

            except errors.ReqlError:
                # the error might have been caused by a comm or temporary error causing a partial batch write

                for row in batch:
                    if table_info.primary_key not in row:
                        raise RuntimeError(
                            "Connection error while importing.  Current row does not have the specified primary key "
                            "(%s), so cannot guarantee absence of duplicates" % table_info.primary_key)
                    res = None
                    if conflict_action == "replace":
                        res = options.retryQuery(
                            "write row to %s.%s" %
                            (db,
                             table),
                            tbl.insert(
                                ast.expr(
                                    row,
                                    nesting_depth=MAX_NESTING_DEPTH),
                                durability=options.durability,
                                conflict=conflict_action,
                                ignore_write_hook=True))
                    else:
                        existingRow = options.retryQuery(
                            "read row from %s.%s" % (db, table),
                            tbl.get(row[table_info.primary_key])
                        )
                        if not existingRow:
                            res = options.retryQuery(
                                "write row to %s.%s" %
                                (db,
                                 table),
                                tbl.insert(
                                    ast.expr(
                                        row,
                                        nesting_depth=MAX_NESTING_DEPTH),
                                    durability=options.durability,
                                    conflict=conflict_action,
                                    ignore_write_hook=True))
                        elif existingRow != row:
                            raise RuntimeError(
                                "Duplicate primary key `%s`:\n%s\n%s" %
                                (table_info.primary_key, str(row), str(existingRow)))

                    if res["errors"] > 0:
                        raise RuntimeError("Error when importing into table '%s.%s': %s" % (
                            db, table, res["first_error"]))
                    if res["inserted"] + res["replaced"] + res["unchanged"] != 1:
                        raise RuntimeError(
                            "The inserted/replaced/unchanged number was not 1 when inserting on '%s.%s': %s" % (
                                db, table, res
                            )
                        )
                    table_info.add_rows_written(1)
            timing_queue.put(('writer_work', time.time() - timePoint))
            timePoint = time.time()

    except Exception as e:
        error_queue.put(Error(str(e), traceback.format_exc(), "%s.%s" % (db, table)))
        exit_event.set()


def update_progress(tables, debug, exit_event, sleep=0.2):
    signal.signal(signal.SIGINT, signal.SIG_IGN)  # workers should not get these

    # give weights to each of the tables based on file size
    totalSize = sum([x.bytes_size for x in tables])
    for table in tables:
        table.weight = float(table.bytes_size) / totalSize

    lastComplete = None
    startTime = time.time()
    readWrites = collections.deque(maxlen=5)  # (time, read, write)
    readWrites.append((startTime, 0, 0))
    readRate = None
    writeRate = None
    while True:
        try:
            if exit_event.is_set():
                break
            complete = read = write = 0
            currentTime = time.time()
            for table in tables:
                complete += table.percent_done * table.weight
                if debug:
                    read += table.rows_read
                    write += table.rows_written
            readWrites.append((currentTime, read, write))
            if complete != lastComplete:
                timeDelta = readWrites[-1][0] - readWrites[0][0]
                if debug and len(readWrites) > 1 and timeDelta > 0:
                    readRate = max((readWrites[-1][1] - readWrites[0][1]) / timeDelta, 0)
                    writeRate = max((readWrites[-1][2] - readWrites[0][2]) / timeDelta, 0)
                utils_common.print_progress(complete, indent=2, read=readRate, write=writeRate)
                lastComplete = complete
            time.sleep(sleep)
        except KeyboardInterrupt:
            break
        except Exception as e:
            if debug:
                print(e)
                traceback.print_exc()


def import_tables(options, sources, files_ignored=None):
    # Make sure this isn't a pre-`reql_admin` cluster - which could result in data loss
    # if the user has a database named 'rethinkdb'
    utils_common.check_minimum_version(options, "1.6")

    start_time = time.time()

    tables = dict(((x.db, x.table), x) for x in sources)  # (db, table) => table

    if six.PY3:
        ctx = multiprocessing.get_context(multiprocessing.get_start_method())
        error_queue = SimpleQueue(ctx=ctx)
        warning_queue = SimpleQueue(ctx=ctx)
        timing_queue = SimpleQueue(ctx=ctx)
    else:
        error_queue = SimpleQueue()
        warning_queue = SimpleQueue()
        timing_queue = SimpleQueue()

    max_queue_size = options.clients * 3
    work_queue = multiprocessing.Manager().Queue(max_queue_size)

    exit_event = multiprocessing.Event()
    interrupt_event = multiprocessing.Event()

    errors = []
    warnings = []
    timing_sums = {}

    pools = []
    progress_bar = None
    progress_bar_sleep = 0.2

    # - setup KeyboardInterupt handler
    signal.signal(signal.SIGINT, lambda a, b: utils_common.abort(pools, exit_event))

    # - queue draining
    def drain_queues():
        # error_queue
        while not error_queue.empty():
            errors.append(error_queue.get())

        # warning_queue
        while not warning_queue.empty():
            warnings.append(warning_queue.get())

        # timing_queue
        while not timing_queue.empty():
            key, value = timing_queue.get()
            if key not in timing_sums:
                timing_sums[key] = value
            else:
                timing_sums[key] += value

    # - setup dbs and tables

    # create missing dbs
    needed_dbs = set([x.db for x in sources])
    if "rethinkdb" in needed_dbs:
        raise RuntimeError("Error: Cannot import tables into the system database: 'rethinkdb'")
    options.retryQuery(
        "ensure dbs: %s" %
        ", ".join(needed_dbs),
        ast.expr(needed_dbs).set_difference(
            query.db_list()).for_each(
            query.db_create(
                query.row)))

    # check for existing tables, or if --force is enabled ones with mis-matched primary keys
    existing_tables = dict([((x["db"], x["name"]), x["primary_key"]) for x in options.retryQuery(
        "list tables", query.db("rethinkdb").table("table_config").pluck(["db", "name", "primary_key"]))])
    already_exist = []
    for source in sources:
        if (source.db, source.table) in existing_tables:
            if not options.force:
                already_exist.append("%s.%s" % (source.db, source.table))
            elif source.primary_key is None:
                source.primary_key = existing_tables[(source.db, source.table)]
            elif source.primary_key != existing_tables[(source.db, source.table)]:
                raise RuntimeError(
                    "Error: Table '%s.%s' already exists with a different primary key: %s (expected: %s)" % (
                        source.db, source.table, existing_tables[(source.db, source.table)], source.primary_key
                    )
                )

    if len(already_exist) == 1:
        raise RuntimeError(
            "Error: Table '%s' already exists, run with --force to import into the existing table" %
            already_exist[0])
    elif len(already_exist) > 1:
        already_exist.sort()
        raise RuntimeError(
            "Error: The following tables already exist, run with --force to import into the existing tables:\n  %s" %
            "\n  ".join(already_exist))

    # - start the import

    try:
        # - start the progress bar
        if not options.quiet:
            progress_bar = multiprocessing.Process(
                target=update_progress,
                name="progress bar",
                args=(sources, options.debug, exit_event, progress_bar_sleep)
            )
            progress_bar.start()
            pools.append([progress_bar])

        # - start the writers
        writers = []
        pools.append(writers)
        for i in range(options.clients):
            writer = multiprocessing.Process(
                target=table_writer,
                name="table writer %d" %
                i,
                kwargs={
                    "tables": tables,
                    "options": options,
                    "work_queue": work_queue,
                    "error_queue": error_queue,
                    "warning_queue": warning_queue,
                    "timing_queue": timing_queue,
                    "exit_event": exit_event})
            writers.append(writer)
            writer.start()

        # - read the tables options.clients at a time
        readers = []
        pools.append(readers)
        file_iter = iter(sources)
        try:
            while not exit_event.is_set():
                # add a workers to fill up the readers pool
                while len(readers) < options.clients:
                    table = next(file_iter)
                    reader = multiprocessing.Process(
                        target=table.read_to_queue,
                        name="table reader %s.%s" %
                        (table.db,
                         table.table),
                        kwargs={
                            "fields": options.fields,
                            "batch_size": options.batch_size,
                            "work_queue": work_queue,
                            "error_queue": error_queue,
                            "warning_queue": warning_queue,
                            "timing_queue": timing_queue,
                            "exit_event": exit_event})
                    readers.append(reader)
                    reader.start()

                # drain the queues
                drain_queues()

                # reap completed tasks
                for reader in readers[:]:
                    if not reader.is_alive():
                        readers.remove(reader)
                    if len(readers) == options.clients:
                        time.sleep(.05)
        except StopIteration:
            pass  # ran out of new tables

        # - wait for the last batch of readers to complete
        while readers:
            # drain the queues
            drain_queues()

            # drain the work queue to prevent readers from stalling on exit
            if exit_event.is_set():
                try:
                    while True:
                        work_queue.get(timeout=0.1)
                except Empty:
                    pass

            # watch the readers
            for reader in readers[:]:
                try:
                    reader.join(.1)
                except Exception as exc:
                    default_logger.exception(exc)
                if not reader.is_alive():
                    readers.remove(reader)

        # - append enough StopIterations to signal all writers
        for _ in writers:
            while True:
                if exit_event.is_set():
                    break
                try:
                    work_queue.put((None, None, StopIteration()), timeout=0.1)
                    break
                except Full:
                    pass

        # - wait for all of the writers
        for writer in writers[:]:
            while writer.is_alive():
                writer.join(0.1)
            writers.remove(writer)

        # - stop the progress bar
        if progress_bar:
            progress_bar.join(progress_bar_sleep * 2)
            if not interrupt_event.is_set():
                utils_common.print_progress(1, indent=2)
            if progress_bar.is_alive():
                progress_bar.terminate()

        # - drain queues
        drain_queues()

        # - final reporting
        if not options.quiet:
            # if successful, make sure 100% progress is reported
            if len(errors) == 0 and not interrupt_event.is_set():
                utils_common.print_progress(1.0, indent=2)

            # advance past the progress bar
            print('')

            # report statistics
            def plural(num, text):
                return "%d %s%s" % (num, text, "" if num == 1 else "s")

            print("  %s imported to %s in %.2f secs" % (plural(sum(x.rows_written for x in sources), "row"),
                                                        plural(len(sources), "table"), time.time() - start_time))

            # report debug statistics
            if options.debug:
                print('Debug timing:')
                for key, value in sorted(timing_sums.items(), key=lambda x: x[0]):
                    print('  %s: %.2f' % (key, value))
    finally:
        signal.signal(signal.SIGINT, signal.SIG_DFL)

    drain_queues()

    for error in errors:
        print("%s" % error.message, file=sys.stderr)
        if options.debug and error.traceback:
            print("  Traceback:\n%s" % error.traceback, file=sys.stderr)
        if len(error.file) == 4:
            print("  In file: %s" % error.file, file=sys.stderr)

    for warning in warnings:
        print("%s" % warning[1], file=sys.stderr)
        if options.debug:
            print("%s traceback: %s" % (warning[0].__name__, warning[2]), file=sys.stderr)
        if len(warning) == 4:
            print("In file: %s" % warning[3], file=sys.stderr)

    if interrupt_event.is_set():
        raise RuntimeError("Interrupted")
    if errors:
        raise RuntimeError("Errors occurred during import")
    if warnings:
        raise RuntimeError("Warnings occurred during import")


def parse_sources(options, files_ignored=None):

    def parse_info_file(path):
        primary_key = None
        indexes = []
        write_hook = None
        with open(path, 'r') as info_file:
            metadata = json.load(info_file)
            if "primary_key" in metadata:
                primary_key = metadata["primary_key"]
            if "indexes" in metadata and options.indexes is not False:
                indexes = metadata["indexes"]
            if "write_hook" in metadata:
                write_hook = metadata["write_hook"]
        return primary_key, indexes, write_hook

    has_write_hooks = utils_common.check_minimum_version(options, '2.3.7', False)

    sources = set()
    if files_ignored is None:
        files_ignored = []
    if options.directory and options.file:
        raise RuntimeError("Error: Both --directory and --file cannot be specified together")
    elif options.file:
        db, table = options.import_table
        path, ext = os.path.splitext(options.file)
        table_type_options = None
        if ext == ".json":
            table_type = JsonSourceFile
        elif ext == ".csv":
            table_type = CsvSourceFile
            table_type_options = {
                'no_header_row': options.no_header,
                'custom_header': options.custom_header
            }
        else:
            raise Exception("The table type is not recognised: %s" % ext)

        # - parse the info file if it exists
        primary_key = options.create_args.get('primary_key', None) if options.create_args else None
        indexes = []
        write_hook = None
        info_path = path + ".info"
        if (primary_key is None or options.indexes is not False) and os.path.isfile(info_path):
            info_primary_key, info_indexes, info_write_hook = parse_info_file(info_path)
            if primary_key is None:
                primary_key = info_primary_key
            if options.indexes is not False:
                indexes = info_indexes
            if write_hook is None:
                write_hook = info_write_hook
            if write_hook and not has_write_hooks:
                raise Exception('this RDB version doesn\'t support write-hooks')

        sources.add(
            table_type(
                source=options.file,
                db=db, table=table,
                query_runner=options.retryQuery,
                primary_key=primary_key,
                indexes=indexes,
                write_hook=write_hook,
                source_options=table_type_options
            )
        )
    elif options.directory:
        # Scan for all files, make sure no duplicated tables with different formats
        dbs = False
        files_ignored = []
        for root, dirs, files in os.walk(options.directory):
            if not dbs:
                files_ignored.extend([os.path.join(root, f) for f in files])
                # The first iteration through should be the top-level directory, which contains the db folders
                dbs = True

                # don't recurse into folders not matching our filter
                db_filter = set([db_table[0] for db_table in options.db_tables or []])
                if db_filter:
                    for dir_name in dirs[:]:  # iterate on a copy
                        if dir_name not in db_filter:
                            dirs.remove(dir_name)
            else:
                if dirs:
                    files_ignored.extend([os.path.join(root, d) for d in dirs])
                    del dirs[:]

                db = os.path.basename(root)
                for filename in files:
                    path = os.path.join(root, filename)
                    table, ext = os.path.splitext(filename)
                    table = os.path.basename(table)

                    if ext not in [".json", ".csv", ".info"]:
                        files_ignored.append(os.path.join(root, filename))
                    elif ext == ".info":
                        pass  # Info files are included based on the data files
                    elif not os.path.exists(os.path.join(root, table + ".info")):
                        files_ignored.append(os.path.join(root, filename))
                    else:
                        # apply db/table filters
                        if options.db_tables:
                            for filter_db, filter_table in options.db_tables:
                                if db == filter_db and filter_table in (None, table):
                                    break  # either all tables in this db, or specific pair
                            else:
                                files_ignored.append(os.path.join(root, filename))
                                continue  # not a chosen db/table

                        # collect the info
                        primary_key = None
                        indexes = []
                        write_hook = None
                        info_path = os.path.join(root, table + ".info")
                        if not os.path.isfile(info_path):
                            files_ignored.append(os.path.join(root, filename))
                        else:
                            primary_key, indexes, write_hook = parse_info_file(info_path)
                        if write_hook and not has_write_hooks:
                            raise Exception('RDB versions below doesn\'t support write-hooks')

                        table_type = None
                        if ext == ".json":
                            table_type = JsonSourceFile
                        elif ext == ".csv":
                            table_type = CsvSourceFile
                        else:
                            raise Exception("The table type is not recognised: %s" % ext)
                        source = table_type(
                            source=path,
                            query_runner=options.retryQuery,
                            db=db, table=table,
                            primary_key=primary_key,
                            indexes=indexes,
                            write_hook=write_hook
                        )

                        # ensure we don't have a duplicate
                        if table in sources:
                            raise RuntimeError(
                                "Error: Duplicate db.table found in directory tree: %s.%s" %
                                (source.db, source.table))

                        sources.add(source)

        # Warn the user about the files that were ignored
        if len(files_ignored) > 0:
            print("Unexpected files found in the specified directory.  Importing a directory expects", file=sys.stderr)
            print(" a directory from `rethinkdb export`.  If you want to import individual tables", file=sys.stderr)
            print(" import them as single files.  The following files were ignored:", file=sys.stderr)
            for ignored_file in files_ignored:
                print("%s" % str(ignored_file), file=sys.stderr)
    else:
        raise RuntimeError("Error: Neither --directory or --file specified")

    return sources


def main(argv=None, prog=None):
    start_time = time.time()

    if argv is None:
        argv = sys.argv[1:]
    options = parse_options(argv, prog=prog)

    try:
        sources = parse_sources(options)
        import_tables(options, sources)
    except RuntimeError as ex:
        print(ex, file=sys.stderr)
        if str(ex) == "Warnings occurred during import":
            return 2
        return 1
    if not options.quiet:
        print("  Done (%d seconds)" % (time.time() - start_time))
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

'''Dispatcher for interactive functions such as repl and backup'''

import code
import sys
import traceback

from rethinkdb import errors, net, utils_common


def startInterpreter(argv=None, prog=None):
    repl_variables = {'r': net.Connection._r, 'rethinkdb': net.Connection._r}
    banner = 'The RethinkDB driver has been imported as `r`.'

    # -- get host/port setup

    # - parse command line
    parser = utils_common.CommonOptionsParser(
        prog=prog, description='An interactive Python shell (repl) with the RethinkDB driver imported')
    options, args = parser.parse_args(argv if argv is not None else sys.argv[1:], connect=False)

    if args:
        parser.error('No positional arguments supported. Unrecognized option(s): %s' % args)

    # -- open connection

    try:
        repl_variables['conn'] = options.retryQuery.conn()
        repl_variables['conn'].repl()
        banner += '''
    A connection to %s:%d has been established as `conn`
    and can be used by calling `run()` on a query without any arguments.''' % (options.hostname, options.driver_port)
    except errors.ReqlDriverError as e:
        banner += '\nWarning: %s' % str(e)
        if options.debug:
            banner += '\n' + traceback.format_exc()

    # -- start interpreter

    code.interact(banner=banner + '\n==========', local=repl_variables)


if __name__ == '__main__':
    if __package__ is None:
        __package__ = 'rethinkdb'

    # -- figure out which mode we are in
    modes = ['dump', 'export', 'import', 'index_rebuild', 'repl', 'restore']

    if len(sys.argv) < 2 or sys.argv[1] not in modes:
        sys.exit('ERROR: Must be called with one of the following verbs: %s' % ', '.join(modes))

    verb = sys.argv[1]
    prog = 'python -m rethinkdb'
    if sys.version_info < (2, 7) or (sys.version_info >= (3, 0) and sys.version_info < (3, 4)):
        prog += '.__main__'  # Python versions 2.6, 3.0, 3.1 and 3.3 do not support running packages
    prog += ' ' + verb
    argv = sys.argv[2:]

    if verb == 'dump':
        from . import _dump
        exit(_dump.main(argv, prog=prog))
    elif verb == 'export':
        from . import _export
        exit(_export.main(argv, prog=prog))
    elif verb == 'import':
        from . import _import
        exit(_import.main(argv, prog=prog))
    elif verb == 'index_rebuild':
        from . import _index_rebuild
        exit(_index_rebuild.main(argv, prog=prog))
    elif verb == 'repl':
        startInterpreter(argv, prog=prog)
    elif verb == 'restore':
        from . import _restore
        exit(_restore.main(argv, prog=prog))



# Copyright 2010-2014 RethinkDB, all rights reserved.

__all__ = [
    'js', 'http', 'json', 'args', 'error', 'random', 'do', 'row', 'branch',
    'union', 'map', 'object', 'binary', 'uuid', 'type_of', 'info', 'range',
    'literal', 'asc', 'desc',
    'db', 'db_create', 'db_drop', 'db_list',
    'table', 'table_create', 'table_drop', 'table_list', 'grant',
    'group', 'reduce', 'count', 'sum', 'avg', 'min', 'max', 'distinct',
    'contains', 'eq', 'ne', 'le', 'ge', 'lt', 'gt', 'and_', 'or_', 'not_',
    'add', 'sub', 'mul', 'div', 'mod', 'bit_and', 'bit_or', 'bit_xor',
    'bit_not', 'bit_sal', 'bit_sar', 'floor', 'ceil',
    'round',
    'time', 'iso8601', 'epoch_time', 'now', 'make_timezone',
    'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday',
    'sunday',
    'january', 'february', 'march', 'april', 'may', 'june',
    'july', 'august', 'september', 'october', 'november', 'december',
    'minval', 'maxval',
    'geojson', 'point', 'line', 'polygon', 'distance', 'intersects', 'circle'
]

from . import ast, ql2_pb2

#
# All top level functions defined here are the starting points for RQL queries
#


def json(*args):
    return ast.Json(*args)


def js(*args, **kwargs):
    return ast.JavaScript(*args, **kwargs)


def args(*args):
    return ast.Args(*args)


def http(url, **kwargs):
    return ast.Http(ast.func_wrap(url), **kwargs)


def error(*msg):
    return ast.UserError(*msg)


def random(*args, **kwargs):
    return ast.Random(*args, **kwargs)


def do(*args):
    return ast.FunCall(*args)


row = ast.ImplicitVar()


def table(*args, **kwargs):
    return ast.Table(*args, **kwargs)


def db(*args):
    return ast.DB(*args)


def db_create(*args):
    return ast.DbCreate(*args)


def db_drop(*args):
    return ast.DbDrop(*args)


def db_list(*args):
    return ast.DbList(*args)


def db_config(*args):
    return ast.DbConfig(*args)


def table_create(*args, **kwargs):
    return ast.TableCreateTL(*args, **kwargs)


def table_drop(*args):
    return ast.TableDropTL(*args)


def table_list(*args):
    return ast.TableListTL(*args)


def grant(*args, **kwargs):
    return ast.GrantTL(*args, **kwargs)


def branch(*args):
    return ast.Branch(*args)


def union(*args):
    return ast.Union(*args)


def map(*args):
    if len(args) > 0:
        # `func_wrap` only the last argument
        return ast.Map(*(args[:-1] + (ast.func_wrap(args[-1]), )))
    else:
        return ast.Map()

# aggregation


def group(*args):
    return ast.Group(*[ast.func_wrap(arg) for arg in args])


def reduce(*args):
    return ast.Reduce(*[ast.func_wrap(arg) for arg in args])


def count(*args):
    return ast.Count(*[ast.func_wrap(arg) for arg in args])


def sum(*args):
    return ast.Sum(*[ast.func_wrap(arg) for arg in args])


def avg(*args):
    return ast.Avg(*[ast.func_wrap(arg) for arg in args])


def min(*args):
    return ast.Min(*[ast.func_wrap(arg) for arg in args])


def max(*args):
    return ast.Max(*[ast.func_wrap(arg) for arg in args])


def distinct(*args):
    return ast.Distinct(*[ast.func_wrap(arg) for arg in args])


def contains(*args):
    return ast.Contains(*[ast.func_wrap(arg) for arg in args])


# orderBy orders
def asc(*args):
    return ast.Asc(*[ast.func_wrap(arg) for arg in args])


def desc(*args):
    return ast.Desc(*[ast.func_wrap(arg) for arg in args])


# math and logic
def eq(*args):
    return ast.Eq(*args)


def ne(*args):
    return ast.Ne(*args)


def lt(*args):
    return ast.Lt(*args)


def le(*args):
    return ast.Le(*args)


def gt(*args):
    return ast.Gt(*args)


def ge(*args):
    return ast.Ge(*args)


def add(*args):
    return ast.Add(*args)


def sub(*args):
    return ast.Sub(*args)


def mul(*args):
    return ast.Mul(*args)


def div(*args):
    return ast.Div(*args)


def mod(*args):
    return ast.Mod(*args)


def bit_and(*args):
    return ast.BitAnd(*args)


def bit_or(*args):
    return ast.BitOr(*args)


def bit_xor(*args):
    return ast.BitXor(*args)


def bit_not(*args):
    return ast.BitNot(*args)


def bit_sal(*args):
    return ast.BitSal(*args)


def bit_sar(*args):
    return ast.BitSar(*args)


def floor(*args):
    return ast.Floor(*args)


def ceil(*args):
    return ast.Ceil(*args)


def round(*args):
    return ast.Round(*args)


def not_(*args):
    return ast.Not(*args)


def and_(*args):
    return ast.And(*args)


def or_(*args):
    return ast.Or(*args)


def type_of(*args):
    return ast.TypeOf(*args)


def info(*args):
    return ast.Info(*args)


def binary(data):
    return ast.Binary(data)


def range(*args):
    return ast.Range(*args)


def time(*args):
    return ast.Time(*args)


def iso8601(*args, **kwargs):
    return ast.ISO8601(*args, **kwargs)


def epoch_time(*args):
    return ast.EpochTime(*args)


def now(*args):
    return ast.Now(*args)


class RqlConstant(ast.RqlQuery):
    def __init__(self, statement, term_type):
        self.statement = statement
        self.term_type = term_type
        super(RqlConstant, self).__init__()

    def compose(self, args, optargs):
        return 'r.' + self.statement


# Time enum values
monday = RqlConstant('monday', ql2_pb2.Term.TermType.MONDAY)
tuesday = RqlConstant('tuesday', ql2_pb2.Term.TermType.TUESDAY)
wednesday = RqlConstant('wednesday', ql2_pb2.Term.TermType.WEDNESDAY)
thursday = RqlConstant('thursday', ql2_pb2.Term.TermType.THURSDAY)
friday = RqlConstant('friday', ql2_pb2.Term.TermType.FRIDAY)
saturday = RqlConstant('saturday', ql2_pb2.Term.TermType.SATURDAY)
sunday = RqlConstant('sunday', ql2_pb2.Term.TermType.SUNDAY)

january = RqlConstant('january', ql2_pb2.Term.TermType.JANUARY)
february = RqlConstant('february', ql2_pb2.Term.TermType.FEBRUARY)
march = RqlConstant('march', ql2_pb2.Term.TermType.MARCH)
april = RqlConstant('april', ql2_pb2.Term.TermType.APRIL)
may = RqlConstant('may', ql2_pb2.Term.TermType.MAY)
june = RqlConstant('june', ql2_pb2.Term.TermType.JUNE)
july = RqlConstant('july', ql2_pb2.Term.TermType.JULY)
august = RqlConstant('august', ql2_pb2.Term.TermType.AUGUST)
september = RqlConstant('september', ql2_pb2.Term.TermType.SEPTEMBER)
october = RqlConstant('october', ql2_pb2.Term.TermType.OCTOBER)
november = RqlConstant('november', ql2_pb2.Term.TermType.NOVEMBER)
december = RqlConstant('december', ql2_pb2.Term.TermType.DECEMBER)

minval = RqlConstant('minval', ql2_pb2.Term.TermType.MINVAL)
maxval = RqlConstant('maxval', ql2_pb2.Term.TermType.MAXVAL)


def make_timezone(*args):
    return ast.RqlTzinfo(*args)


# Merge values
def literal(*args):
    return ast.Literal(*args)


def object(*args):
    return ast.Object(*args)


def uuid(*args):
    return ast.UUID(*args)


# Global geospatial operations
def geojson(*args):
    return ast.GeoJson(*args)


def point(*args):
    return ast.Point(*args)


def line(*args):
    return ast.Line(*args)


def polygon(*args):
    return ast.Polygon(*args)


def distance(*args, **kwargs):
    return ast.Distance(*args, **kwargs)


def intersects(*args):
    return ast.Intersects(*args)


def circle(*args, **kwargs):
    return ast.Circle(*args, **kwargs)



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

import asyncio
import contextlib
import socket
import ssl
import struct

from rethinkdb import ql2_pb2
from rethinkdb.errors import ReqlAuthError, ReqlCursorEmpty, ReqlDriverError, ReqlTimeoutError, RqlCursorEmpty
from rethinkdb.net import Connection as ConnectionBase, Cursor, Query, Response, maybe_profile


__all__ = ['Connection']


pResponse = ql2_pb2.Response.ResponseType
pQuery = ql2_pb2.Query.QueryType


@asyncio.coroutine
def _read_until(streamreader, delimiter):
    """Naive implementation of reading until a delimiter"""
    buffer = bytearray()

    while True:
        c = (yield from streamreader.read(1))
        if c == b'':
            break  # EOF
        buffer.append(c[0])
        if c == delimiter:
            break

    return bytes(buffer)


def reusable_waiter(loop, timeout):
    """Wait for something, with a timeout from when the waiter was created.

    This can be used in loops::

        waiter = reusable_waiter(event_loop, 10.0)
        while some_condition:
            yield from waiter(some_future)
    """
    if timeout is not None:
        deadline = loop.time() + timeout
    else:
        deadline = None

    @asyncio.coroutine
    def wait(future):
        if deadline is not None:
            new_timeout = max(deadline - loop.time(), 0)
        else:
            new_timeout = None
        return (yield from asyncio.wait_for(future, new_timeout, loop=loop))

    return wait


@contextlib.contextmanager
def translate_timeout_errors():
    try:
        yield
    except asyncio.TimeoutError:
        raise ReqlTimeoutError()


# The asyncio implementation of the Cursor object:
# The `new_response` Future notifies any waiting coroutines that the can attempt
# to grab the next result.  In addition, the waiting coroutine will schedule a
# timeout at the given deadline (if provided), at which point the future will be
# errored.
class AsyncioCursor(Cursor):
    def __init__(self, *args, **kwargs):
        Cursor.__init__(self, *args, **kwargs)
        self.new_response = asyncio.Future()

    def __aiter__(self):
        return self

    @asyncio.coroutine
    def __anext__(self):
        try:
            return (yield from self._get_next(None))
        except ReqlCursorEmpty:
            raise StopAsyncIteration

    @asyncio.coroutine
    def close(self):
        if self.error is None:
            self.error = self._empty_error()
            if self.conn.is_open():
                self.outstanding_requests += 1
                yield from self.conn._parent._stop(self)

    def _extend(self, res_buf):
        Cursor._extend(self, res_buf)
        self.new_response.set_result(True)
        self.new_response = asyncio.Future()

    # Convenience function so users know when they've hit the end of the cursor
    # without having to catch an exception
    @asyncio.coroutine
    def fetch_next(self, wait=True):
        timeout = Cursor._wait_to_timeout(wait)
        waiter = reusable_waiter(self.conn._io_loop, timeout)
        while len(self.items) == 0 and self.error is None:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            with translate_timeout_errors():
                yield from waiter(asyncio.shield(self.new_response))
        # If there is a (non-empty) error to be received, we return True, so the
        # user will receive it on the next `next` call.
        return len(self.items) != 0 or not isinstance(self.error, RqlCursorEmpty)

    def _empty_error(self):
        # We do not have RqlCursorEmpty inherit from StopIteration as that interferes
        # with mechanisms to return from a coroutine.
        return RqlCursorEmpty()

    @asyncio.coroutine
    def _get_next(self, timeout):
        waiter = reusable_waiter(self.conn._io_loop, timeout)
        while len(self.items) == 0:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            with translate_timeout_errors():
                yield from waiter(asyncio.shield(self.new_response))
        return self.items.popleft()

    def _maybe_fetch_batch(self):
        if self.error is None and \
           len(self.items) < self.threshold and \
           self.outstanding_requests == 0:
            self.outstanding_requests += 1
            asyncio.ensure_future(self.conn._parent._continue(self))


class ConnectionInstance(object):
    _streamreader = None
    _streamwriter = None
    _reader_task = None

    def __init__(self, parent, io_loop=None):
        self._parent = parent
        self._closing = False
        self._user_queries = {}
        self._cursor_cache = {}
        self._ready = asyncio.Future()
        self._io_loop = io_loop
        if self._io_loop is None:
            self._io_loop = asyncio.get_event_loop()

    def client_port(self):
        if self.is_open():
            return self._streamwriter.get_extra_info('sockname')[1]

    def client_address(self):
        if self.is_open():
            return self._streamwriter.get_extra_info('sockname')[0]

    @asyncio.coroutine
    def connect(self, timeout):
        try:
            ssl_context = None
            if len(self._parent.ssl) > 0:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                if hasattr(ssl_context, "options"):
                    ssl_context.options |= getattr(ssl, "OP_NO_SSLv2", 0)
                    ssl_context.options |= getattr(ssl, "OP_NO_SSLv3", 0)
                ssl_context.verify_mode = ssl.CERT_REQUIRED
                ssl_context.check_hostname = True  # redundant with match_hostname
                ssl_context.load_verify_locations(self._parent.ssl["ca_certs"])

            self._streamreader, self._streamwriter = yield from asyncio.open_connection(
                self._parent.host, self._parent.port, loop=self._io_loop, ssl=ssl_context
            )
            self._streamwriter.get_extra_info('socket').setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._streamwriter.get_extra_info('socket').setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        except Exception as err:
            raise ReqlDriverError('Could not connect to %s:%s. Error: %s' %
                                  (self._parent.host, self._parent.port, str(err)))

        try:
            self._parent.handshake.reset()
            response = None
            with translate_timeout_errors():
                while True:
                    request = self._parent.handshake.next_message(response)
                    if request is None:
                        break
                    # This may happen in the `V1_0` protocol where we send two requests as
                    # an optimization, then need to read each separately
                    if request is not "":
                        self._streamwriter.write(request)

                    response = yield from asyncio.wait_for(
                        _read_until(self._streamreader, b'\0'),
                        timeout, loop=self._io_loop,
                    )
                    response = response[:-1]
        except ReqlAuthError:
            yield from self.close()
            raise
        except ReqlTimeoutError as err:
            yield from self.close()
            raise ReqlDriverError(
                'Connection interrupted during handshake with %s:%s. Error: %s' % (
                    self._parent.host, self._parent.port, str(err)
                )
            )
        except Exception as err:
            yield from self.close()
            raise ReqlDriverError('Could not connect to %s:%s. Error: %s' %
                                  (self._parent.host, self._parent.port, str(err)))

        # Start a parallel function to perform reads
        #  store a reference to it so it doesn't get destroyed
        self._reader_task = asyncio.ensure_future(self._reader(), loop=self._io_loop)
        return self._parent

    def is_open(self):
        return not (self._closing or self._streamreader.at_eof())

    @asyncio.coroutine
    def close(self, noreply_wait=False, token=None, exception=None):
        self._closing = True
        if exception is not None:
            err_message = "Connection is closed (%s)." % str(exception)
        else:
            err_message = "Connection is closed."

        # Cursors may remove themselves when errored, so copy a list of them
        for cursor in list(self._cursor_cache.values()):
            cursor._error(err_message)

        for query, future in iter(self._user_queries.values()):
            if not future.done():
                future.set_exception(ReqlDriverError(err_message))

        self._user_queries = {}
        self._cursor_cache = {}

        if noreply_wait:
            noreply = Query(pQuery.NOREPLY_WAIT, token, None, None)
            yield from self.run_query(noreply, False)

        self._streamwriter.close()
        # We must not wait for the _reader_task if we got an exception, because that
        # means that we were called from it. Waiting would lead to a deadlock.
        if self._reader_task and exception is None:
            yield from self._reader_task

        return None

    @asyncio.coroutine
    def run_query(self, query, noreply):
        self._streamwriter.write(query.serialize(self._parent._get_json_encoder(query)))
        if noreply:
            return None

        response_future = asyncio.Future()
        self._user_queries[query.token] = (query, response_future)
        return (yield from response_future)

    # The _reader coroutine runs in parallel, reading responses
    # off of the socket and forwarding them to the appropriate Future or Cursor.
    # This is shut down as a consequence of closing the stream, or an error in the
    # socket/protocol from the server.  Unexpected errors in this coroutine will
    # close the ConnectionInstance and be passed to any open Futures or Cursors.
    @asyncio.coroutine
    def _reader(self):
        try:
            while True:
                buf = yield from self._streamreader.readexactly(12)
                (token, length,) = struct.unpack("<qL", buf)
                buf = yield from self._streamreader.readexactly(length)

                cursor = self._cursor_cache.get(token)
                if cursor is not None:
                    cursor._extend(buf)
                elif token in self._user_queries:
                    # Do not pop the query from the dict until later, so
                    # we don't lose track of it in case of an exception
                    query, future = self._user_queries[token]
                    res = Response(token, buf,
                                   self._parent._get_json_decoder(query))
                    if res.type == pResponse.SUCCESS_ATOM:
                        future.set_result(maybe_profile(res.data[0], res))
                    elif res.type in (pResponse.SUCCESS_SEQUENCE,
                                      pResponse.SUCCESS_PARTIAL):
                        cursor = AsyncioCursor(self, query, res)
                        future.set_result(maybe_profile(cursor, res))
                    elif res.type == pResponse.WAIT_COMPLETE:
                        future.set_result(None)
                    elif res.type == pResponse.SERVER_INFO:
                        future.set_result(res.data[0])
                    else:
                        future.set_exception(res.make_error(query))
                    del self._user_queries[token]
                elif not self._closing:
                    raise ReqlDriverError("Unexpected response received.")
        except Exception as ex:
            if not self._closing:
                yield from self.close(exception=ex)


class Connection(ConnectionBase):
    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(ConnectionInstance, *args, **kwargs)
        try:
            self.port = int(self.port)
        except ValueError:
            raise ReqlDriverError("Could not convert port %s to an integer." % self.port)

    @asyncio.coroutine
    def __aenter__(self):
        return self

    @asyncio.coroutine
    def __aexit__(self, exception_type, exception_val, traceback):
        yield from self.close(False)

    @asyncio.coroutine
    def _stop(self, cursor):
        self.check_open()
        q = Query(pQuery.STOP, cursor.query.token, None, None)
        return (yield from self._instance.run_query(q, True))

    @asyncio.coroutine
    def reconnect(self, noreply_wait=True, timeout=None):
        # We close before reconnect so reconnect doesn't try to close us
        # and then fail to return the Future (this is a little awkward).
        yield from self.close(noreply_wait)
        self._instance = self._conn_type(self, **self._child_kwargs)
        return (yield from self._instance.connect(timeout))

    @asyncio.coroutine
    def close(self, noreply_wait=True):
        if self._instance is None:
            return None
        return (yield from ConnectionBase.close(self, noreply_wait=noreply_wait))



# Copyright 2019 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2019 RethinkDB all rights reserved.

import collections
import contextlib
import socket
import ssl
import struct

import trio
import trio.abc

from rethinkdb import ql2_pb2, RethinkDB
from rethinkdb.errors import ReqlAuthError, ReqlCursorEmpty, ReqlDriverError, \
    ReqlTimeoutError, RqlCursorEmpty
from rethinkdb.net import Connection as ConnectionBase, Cursor, Query, \
    Response, maybe_profile, make_connection


__all__ = ['Connection']


P_RESPONSE = ql2_pb2.Response.ResponseType
P_QUERY = ql2_pb2.Query.QueryType


class TrioFuture:
    ''' Trio does not have a future class because Trio encourages the use of
    "coroutines all the way down", but the this driver was implemented by
    copying the net_asyncio code and transliterating it into the Trio API. The
    underlying code in net.py has the I/O intertwined with framing and state
    logic, making it difficult to cleanly write async code in the Trio style.
    Therefore I've taken the easy way out by writing up a simple future class.

    Similar to an asyncio future except without callbacks or cancellation.
    '''
    def __init__(self):
        self._event = trio.Event()
        self._cancelled = False
        self._value = None
        self._exc = None

    async def wait(self):
        await self._event.wait()
        return self.result()

    def exception(self):
        if self._event.is_set():
            return self._exc
        else:
            raise Exception('Future value has not been set')

    def result(self):
        if self._event.is_set():
            if self._exc is not None:
                raise self._exc
            return self._value
        else:
            raise Exception('Future value has not been set')

    def set_result(self, value):
        self._value = value
        self._event.set()

    def set_exception(self, exc):
        self._exc = exc
        self._event.set()

    def done(self):
        return self._event.is_set()


@contextlib.contextmanager
def _reql_timeout(seconds):
    '''
    Run a block with a timeout, raising `ReqlTimeoutError` if the block
    execution exceeds the timeout.

    :param float seconds: A timeout in seconds. If None, then no timeout is
        enforced.
    :raises ReqlTimeoutError: If execution time exceeds the timeout.
    '''
    if seconds is None:
        yield
    else:
        try:
            with trio.fail_after(seconds):
                yield
        except trio.TooSlow:
            raise ReqlTimeoutError()


class TrioCursor(Cursor, trio.abc.AsyncResource):
    ''' A cursor that allows async iteration within the Trio framework. '''
    def __init__(self, *args, **kwargs):
        ''' Constructor '''
        self._new_response = trio.Event()
        self._nursery = kwargs.pop('nursery')
        Cursor.__init__(self, *args, **kwargs)

    def __aiter__(self):
        ''' This object is an async iterator. '''
        return self

    async def __anext__(self):
        ''' Asynchronously get next item from this cursor. '''
        try:
            return await self._get_next(timeout=None)
        except ReqlCursorEmpty:
            raise StopAsyncIteration

    async def close(self):
        ''' Close this cursor. '''
        if self.error is None:
            self.error = self._empty_error()
            if self.conn.is_open():
                self.outstanding_requests += 1
                await self.conn._parent._stop(self)
    aclose = close

    def _extend(self, res_buf):
        ''' Override so that we can make this async, and also to wake up blocked
        tasks. '''
        self.outstanding_requests -= 1
        self._maybe_fetch_batch()
        res = Response(self.query.token, res_buf, self._json_decoder)
        self._extend_internal(res)
        self._new_response.set()
        self._new_response = trio.Event()

    # Convenience function so users know when they've hit the end of the cursor
    # without having to catch an exception
    async def fetch_next(self, wait=True):
        timeout = Cursor._wait_to_timeout(wait)
        while len(self.items) == 0:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            with _reql_timeout(timeout):
                await self._new_response.wait()
        # If there is a (non-empty) error to be received, we return True, so the
        # user will receive it on the next `next` call.
        return len(self.items) != 0

    def _empty_error(self):
        # We do not have RqlCursorEmpty inherit from StopIteration as that interferes
        # with mechanisms to return from a coroutine.
        return RqlCursorEmpty()

    async def _get_next(self, timeout):
        while len(self.items) == 0:
            self._maybe_fetch_batch()
            if self.error is not None:
                raise self.error
            with _reql_timeout(timeout):
                await self._new_response.wait()
        item = self.items.popleft()
        if isinstance(item, Exception):
            raise item
        return item

    async def _parent_continue(self):
        return await self.conn._parent._continue(self)

    def _maybe_fetch_batch(self):
        if self.error is None and len(self.items) < self.threshold and \
                self.outstanding_requests == 0:
            self.outstanding_requests += 1
            self._nursery.start_soon(self.conn._parent._continue, self)


class ConnectionInstance:
    def __init__(self, parent, nursery=None):
        self._stream = None
        self._stream_lock = trio.Lock()
        self._sockname = None
        self._parent = parent
        self._closing = False
        self._closed = False
        self._user_queries = {}
        self._cursor_cache = {}
        self._reader_ended_event = None
        self._nursery = nursery

    def client_port(self):
        if self.is_open():
            return self._sockname[1]

    def client_address(self):
        if self.is_open():
            return self._sockname[0]

    async def _send(self, data):
        async with self._stream_lock:
            try:
                await self._stream.send_all(data)
            except (trio.BrokenResourceError, trio.ClosedResourceError):
                self._closed = True

    async def _read_until(self, delimiter):
        ''' Naive implementation of reading until a delimiter. '''
        buffer = bytearray()

        try:
            while True:
                data = await self._stream.receive_some(1)
                buffer.append(data[0])
                if data == delimiter:
                    break
        except (trio.BrokenResourceError, trio.ClosedResourceError):
            self._closed = True

        return bytes(buffer)

    async def _read_exactly(self, num):
        try:
            return await self._stream.receive_some(num)
        except (trio.BrokenResourceError, trio.ClosedResourceError):
            self._closed = True

    async def connect(self, timeout):
        try:
            ssl_context = None
            if len(self._parent.ssl) > 0:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                if hasattr(ssl_context, "options"):
                    ssl_context.options |= getattr(ssl, "OP_NO_SSLv2", 0)
                    ssl_context.options |= getattr(ssl, "OP_NO_SSLv3", 0)
                ssl_context.verify_mode = ssl.CERT_REQUIRED
                ssl_context.check_hostname = True  # redundant with match_hostname
                ssl_context.load_verify_locations(self._parent.ssl["ca_certs"])
            if ssl_context:
                self._stream = await trio.open_ssl_over_tcp_stream(
                    self._parent.host, self._parent.port,
                    ssl_context=ssl_context)
                socket_ = self._stream.transport_stream.socket
            else:
                self._stream = await trio.open_tcp_stream(self._parent.host,
                    self._parent.port)
                socket_ = self._stream.socket
            self._sockname = socket_.getsockname()
            socket_.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        except Exception as err:
            raise ReqlDriverError('Could not connect to %s:%s. Error: %s' %
                                  (self._parent.host, self._parent.port, str(err)))

        try:
            self._parent.handshake.reset()
            response = None
            while True:
                request = self._parent.handshake.next_message(response)
                if request is None:
                    break
                # This may happen in the `V1_0` protocol where we send two requests as
                # an optimization, then need to read each separately
                if request is not "":
                    await self._send(request)
                with _reql_timeout(timeout):
                    response = await self._read_until(b'\0')
                response = response[:-1]
        except ReqlAuthError:
            await self.close()
            raise
        except ReqlTimeoutError as err:
            await self.close()
            raise ReqlDriverError(
                'Connection interrupted during handshake with %s:%s. Error: %s' % (
                    self._parent.host, self._parent.port, str(err)
                )
            )
        except Exception as err:
            await self.close()
            raise ReqlDriverError('Could not connect to %s:%s. Error: %s' %
                                  (self._parent.host, self._parent.port, str(err)))

        # Start a parallel function to perform reads
        self._nursery.start_soon(self._reader_task)
        return self._parent

    def is_open(self):
        return not (self._closing or self._closed)

    async def close(self, noreply_wait=False, token=None, exception=None):
        self._closing = True
        if exception is not None:
            err_message = "Connection is closed (%s)." % str(exception)
        else:
            err_message = "Connection is closed."

        # Cursors may remove themselves when errored, so copy a list of them
        for cursor in list(self._cursor_cache.values()):
            cursor._error(err_message)

        for _, future in self._user_queries.values():
            if not future.done():
                future.set_exception(ReqlDriverError(err_message))

        self._user_queries = {}
        self._cursor_cache = {}

        if noreply_wait:
            noreply = Query(P_QUERY.NOREPLY_WAIT, token, None, None)
            await self.run_query(noreply, False)

        try:
            await self._stream.aclose()
        except (trio.ClosedResourceError, trio.BrokenResourceError):
            pass
        # We must not wait for the _reader_task if we got an exception, because that
        # means that we were called from it. Waiting would lead to a deadlock.
        if self._reader_ended_event:
            await self._reader_ended_event.wait()

        return None

    async def run_query(self, query, noreply):
        await self._send(query.serialize(self._parent._get_json_encoder(query)))
        if noreply:
            return None

        response_future = TrioFuture()
        self._user_queries[query.token] = (query, response_future)
        return await response_future.wait()

    # The reader task runs in parallel, reading responses
    # off of the socket and forwarding them to the appropriate Future or Cursor.
    # This is shut down as a consequence of closing the stream, or an error in the
    # socket/protocol from the server. Unexpected errors in this coroutine will
    # close the ConnectionInstance and be passed to any open Futures or Cursors.
    async def _reader_task(self):
        self._reader_ended_event = trio.Event()
        try:
            while True:
                buf = await self._read_exactly(12)
                (token, length,) = struct.unpack("<qL", buf)
                buf = await self._read_exactly(length)

                cursor = self._cursor_cache.get(token)
                if cursor is not None:
                    cursor._extend(buf)
                elif token in self._user_queries:
                    # Do not pop the query from the dict until later, so
                    # we don't lose track of it in case of an exception
                    query, future = self._user_queries[token]
                    res = Response(token, buf,
                                   self._parent._get_json_decoder(query))
                    if res.type == P_RESPONSE.SUCCESS_ATOM:
                        future.set_result(maybe_profile(res.data[0], res))
                    elif res.type in (P_RESPONSE.SUCCESS_SEQUENCE,
                                      P_RESPONSE.SUCCESS_PARTIAL):
                        cursor = TrioCursor(self, query, res, nursery=self._nursery)
                        future.set_result(maybe_profile(cursor, res))
                    elif res.type == P_RESPONSE.WAIT_COMPLETE:
                        future.set_result(None)
                    elif res.type == P_RESPONSE.SERVER_INFO:
                        future.set_result(res.data[0])
                    else:
                        future.set_exception(res.make_error(query))
                    del self._user_queries[token]
                elif not self._closing:
                    raise ReqlDriverError("Unexpected response received.")
        except Exception as ex:
            if not self._closing:
                await self.close(exception=ex)
        finally:
            self._reader_ended_event.set()


class Connection(ConnectionBase):
    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(ConnectionInstance, *args, **kwargs)
        try:
            self.port = int(self.port)
        except ValueError:
            raise ReqlDriverError("Could not convert port %s to an integer." % self.port)

    async def _stop(self, cursor):
        self.check_open()
        query = Query(P_QUERY.STOP, cursor.query.token, None, None)
        return await self._instance.run_query(query, True)

    async def reconnect(self, noreply_wait=True, timeout=None):
        await self.close(noreply_wait)
        self._instance = self._conn_type(self, **self._child_kwargs)
        return await self._instance.connect(timeout)

    async def close(self, noreply_wait=True):
        if self._instance is None:
            return None
        return await ConnectionBase.close(self, noreply_wait=noreply_wait)


class AsyncTrioConnectionContextManager:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._conn = None

    @classmethod
    def open(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    async def __aenter__(self):
        self._conn = await make_connection(Connection, *self._args,
            **self._kwargs)
        return self._conn

    async def __aexit__(self, exc_type, exc, traceback):
        await self._conn.close(False)


# Monkeypatch RethinkDB with an async context manager.
RethinkDB.open = AsyncTrioConnectionContextManager.open


class _TrioConnectionPoolContextManager:
    '''
    A context manager for a trio connection pool. This automatically acquires
    a connection from the pool when entering the block, then releases it after
    exiting the block.

    This is not meant to be instantiated directly. Use
    TrioConnectionPool.connection() instead.
    '''
    def __init__(self, pool):
        self._conn = None
        self._pool = pool

    async def __aenter__(self):
        ''' Acquire a connection. '''
        self._conn = await self._pool.acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc, traceback):
        ''' Release a connection. '''
        await self._pool.release(self._conn)


class TrioConnectionPool:
    ''' A RethinkDB connection pool for Trio framework. '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor.

        :param int max_idle: The maximum number of idle connections to keep in
        the pool.
        '''
        self._closed = False
        self._args = args
        self._kwargs = kwargs
        self._nursery = kwargs['nursery']
        self._max_idle = kwargs.pop('maxidle', 10)
        self._connections = collections.deque()
        self._lent_out = set()

    def connection(self):
        return _TrioConnectionPoolContextManager(self)

    async def acquire(self):
        if self._closed:
            raise Exception('DB pool is closed!')

        try:
            conn = self._connections.popleft()
            while not conn.is_open():
                # Connections in the pool may timeout, so look for one that is
                # still connected.
                conn = self._connections.popleft()
        except IndexError:
            conn = await make_connection(*self._args, **self._kwargs)

        self._lent_out.add(conn)
        return conn

    async def release(self, conn):
        self._lent_out.remove(conn)
        if len(self._connections) < self._max_idle:
            self._connections.append(conn)
        else:
            await conn.close()

    async def close(self):
        async with trio.open_nursery() as nursery:
            for conn in self._connections:
                nursery.start_soon(conn.close)
            for conn in self._lent_out:
                nursery.start_soon(conn.close)

        self._closed = True


# Monkeypatch RethinkDB with a connection pool.
RethinkDB.ConnectionPool = TrioConnectionPool
